import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
import matplotlib.pyplot as plt
from scipy import optimize
import seaborn as sns
from patsy import dmatrices, dmatrix, demo_data

from .. import inference as ifr
from .. import families as fam
from .. import output as op
from .. import tsm as tsm
from .. import data_check as dcf
from .. import covariances as cov
from .. import results as res

from .kalman import *
from .dynlin import *

class NDynReg(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** NON-GAUSSIAN DYNAMIC REGRESSION MODELS ****

    Parameters
    ----------

    formula : string
        patsy string describing the regression

    data : pd.DataFrame
        Field to specify the data that will be used
    """

    def __init__(self,formula,data):

        # Initialize TSM object
        super(NDynReg,self).__init__('NDynReg')

        # Latent variables
        self.max_lag = 0
        self._z_hide = 0 # Whether to cutoff variance latent variables from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"
        self.multivariate_model = False

        # Format the data
        self.is_pandas = True # This is compulsory for this model type
        self.data_original = data
        self.formula = formula
        self.y, self.X = dmatrices(formula, data)
        self.z_no = self.X.shape[1]
        self.y_name = self.y.design_info.describe()
        self.data_name = self.y_name
        self.X_names = self.X.design_info.describe().split(" + ")
        self.y = np.array([self.y]).ravel()
        self.data = self.y
        self.X = np.array([self.X])[0]
        self.index = data.index
        self.state_no = self.X.shape[1]

        self._create_latent_variables()

    def _create_latent_variables(self):
        """ Creates model latent variables

        Returns
        ----------
        None (changes model attributes)
        """

        for parm in range(self.z_no):
            self.latent_variables.add_z('Sigma^2 ' + self.X_names[parm], fam.Flat(transform='exp'), fam.Normal(0,3))

    def _get_scale_and_shape(self):
        """ Retrieves the scale and shape for the model

        Returns
        ----------
        Scale (float) and shape (float)
        """
        if self.dist == 't':
            return self.latent_variables.get_z_values(transformed=True)[-2],self.latent_variables.get_z_values(transformed=True)[-1],0
        elif self.dist == 'Laplace':
            return self.latent_variables.get_z_values(transformed=True)[-1],0,0
        elif self.dist == 'skewt':
            return self.latent_variables.get_z_values(transformed=True)[-2],self.latent_variables.get_z_values(transformed=True)[-1],self.latent_variables.get_z_values(transformed=True)[-3]
        else:
            return 0, 0, 0

    def _model(self,data,beta):
        """ Creates the structure of the model

        Parameters
        ----------
        data : np.array
            Contains the time series

        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        a,P,K,F,v : np.array
            Filted states, filtered variances, Kalman gains, F matrix, residuals
        """     

        T, Z, R, Q, H = self._ss_matrices(beta)

        return nld_univariate_kalman(data,Z,H,T,Q,R,0.0)

    def _ss_matrices(self,beta):
        """ Creates the state space matrices required

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm
        """     


        T = np.identity(self.state_no)
        Z = self.X
        R = np.identity(self.state_no)
        
        Q = np.identity(self.state_no)
        for i in range(0,self.state_no):
            Q[i][i] = self.latent_variables.z_list[i].prior.transform(beta[i])

        return T, Z, R, Q

    def _general_approximating_model(self,beta,T,Z,R,Q,h_approx):
        """ Creates simplest approximating model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm

        h_approx : float
            Value to use for the H matrix

        Returns
        ----------

        H : np.array
            Approximating measurement variance matrix

        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(self.data.shape[0])*h_approx
        mu = np.zeros(self.data.shape[0])

        return H, mu

    def _poisson_approximating_model(self,beta,T,Z,R,Q):
        """ Creates approximating Gaussian model for Poisson measurement density

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm

        Returns
        ----------

        H : np.array
            Approximating measurement variance matrix

        mu : np.array
            Approximating measurement constants
        """     

        if hasattr(self, 'H'):
            H = self.H
        else:
            H = np.ones(self.data.shape[0])

        if hasattr(self, 'mu'):
            mu = self.mu
        else:
            mu = np.zeros(self.data.shape[0])

        alpha = np.zeros([self.state_no, self.data.shape[0]])
        tol = 100.0
        it = 0
        while tol > 10**-7 and it < 5:
            old_alpha = np.sum(self.X*alpha.T,axis=1)
            alpha, V = nld_univariate_KFS(self.data,Z,H,T,Q,R,mu)
            H = np.exp(-np.sum(self.X*alpha.T,axis=1))
            mu = self.data - np.sum(self.X*alpha.T,axis=1) - np.exp(-np.sum(self.X*alpha.T,axis=1))*(self.data - np.exp(np.sum(self.X*alpha.T,axis=1)))
            tol = np.mean(np.abs(np.sum(self.X*alpha.T,axis=1)-old_alpha))
            it += 1

        return H, mu

    def _t_approximating_model(self,beta,T,Z,R,Q):
        """ Creates approximating Gaussian model for t measurement density

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm

        Returns
        ----------

        H : np.array
            Approximating measurement variance matrix

        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(self.data.shape[0])*self.latent_variables.z_list[-2].prior.transform(beta[-2])
        mu = np.zeros(self.data.shape[0])

        return H, mu

    def _skewt_approximating_model(self,beta,T,Z,R,Q):
        """ Creates approximating Gaussian model for t measurement density

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm

        Returns
        ----------

        H : np.array
            Approximating measurement variance matrix

        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(self.data.shape[0])*self.latent_variables.z_list[-2].prior.transform(beta[-2])
        mu = np.zeros(self.data.shape[0])

        return H, mu


    @classmethod
    def Exponential(cls,formula,data):
        """ Creates Exponential-distributed state space model

        Parameters
        ----------
        data : np.array
            Contains the time series

        integ : int (default : 0)
            Specifies how many time to difference the time series.

        target : str (pd.DataFrame) or int (np.array)
            Specifies which column name or array index to use. By default, first
            column/array will be selected as the dependent variable.

        Returns
        ----------
        - NDynReg.Exponential object
        """     

        x = NDynReg(formula=formula,data=data)
        x.meas_likelihood = x.exponential_likelihood
        x.model_name = "Exponential Dynamic Regression Model"   
        x.dist = "Exponential"
        x.link = np.exp
        temp = DynReg(formula=formula,data=data)
        temp.fit()

        for i in range(x.z_no):
            x.latent_variables.z_list[i].start = temp.latent_variables.get_z_values()[i+1]

        def approx_model(beta,T,Z,R,Q):
            return x._general_approximating_model(beta,T,Z,R,Q,temp.latent_variables.get_z_values(transformed=True)[0])

        x._approximating_model = approx_model

        def draw_variable(loc,scale,shape,skewness,nsims):
            return np.random.exponential(1/loc, nsims)

        x.m_likelihood_markov_blanket = x.exponential_likelihood_markov_blanket

        return x

    @classmethod
    def Laplace(cls,formula,data):
        """ Creates Laplace-distributed state space model

        Parameters
        ----------
        data : np.array
            Contains the time series

        integ : int (default : 0)
            Specifies how many time to difference the time series.

        target : str (pd.DataFrame) or int (np.array)
            Specifies which column name or array index to use. By default, first
            column/array will be selected as the dependent variable.

        Returns
        ----------
        - NDynReg.Laplace object
        """     

        x = NDynReg(formula=formula,data=data)
        
        x.latent_variables.add_z('Laplace Scale', fam.Flat(transform='exp'), fam.Normal(0,3))
        x.z_no += 1
        x.meas_likelihood = x.laplace_likelihood
        x.model_name = "Laplace Dynamic Regression Model"   
        x.dist = "Laplace"
        x.link = np.array
        temp = DynReg(formula=formula,data=data)
        temp.fit()

        for i in range(x.z_no-1):
            x.latent_variables.z_list[i].start = temp.latent_variables.get_z_values()[i+1]

        x.latent_variables.z_list[-1].start = temp.latent_variables.get_z_values()[0]

        def approx_model(beta,T,Z,R,Q):
            return x._general_approximating_model(beta,T,Z,R,Q,temp.latent_variables.get_z_values(transformed=True)[0])

        x._approximating_model = approx_model

        def draw_variable(loc,scale,shape,skewness,nsims):
            return np.random.laplace(loc, scale, nsims)

        x.draw_variable = draw_variable
        x.m_likelihood_markov_blanket = x.laplace_likelihood_markov_blanket

        return x

    @classmethod
    def Poisson(cls,formula,data):
        """ Creates Poisson-distributed state space model

        Parameters
        ----------
        data : np.array
            Contains the time series

        integ : int (default : 0)
            Specifies how many time to difference the time series.

        target : str (pd.DataFrame) or int (np.array)
            Specifies which column name or array index to use. By default, first
            column/array will be selected as the dependent variable.

        Returns
        ----------
        - NDynReg.Poisson object
        """     

        x = NDynReg(formula=formula,data=data)
        x._approximating_model = x._poisson_approximating_model
        x.meas_likelihood = x.poisson_likelihood
        x.model_name = "Poisson Dynamic Regression Model"   
        x.dist = "Poisson"
        x.link = np.exp
        temp = DynReg(formula=formula,data=data)
        temp.fit()
        for i in range(x.z_no):
            x.latent_variables.z_list[i].start = temp.latent_variables.get_z_values()[i+1]

        def draw_variable(loc,scale,shape,skewness,nsims):
            return np.random.poisson(loc, nsims)

        x.draw_variable = draw_variable
        x.m_likelihood_markov_blanket = x.poisson_likelihood_markov_blanket

        return x

    @classmethod
    def t(cls,formula,data):
        """ Creates t-distributed state space model

        Parameters
        ----------
        data : np.array
            Contains the time series

        integ : int (default : 0)
            Specifies how many time to difference the time series.

        target : str (pd.DataFrame) or int (np.array)
            Specifies which column name or array index to use. By default, first
            column/array will be selected as the dependent variable.

        Returns
        ----------
        - NLLEV.t object
        """     

        x = NDynReg(formula=formula,data=data)
        
        x.latent_variables.add_z('Signal^2 irregular', fam.Flat(transform='exp'), fam.Normal(0,3))
        x.latent_variables.add_z('v', fam.Flat(transform='exp'), fam.Normal(0,3))
        x.z_no += 2

        x._approximating_model = x._t_approximating_model
        x.meas_likelihood = x.t_likelihood
        x.model_name = "t-distributed Dynamic Regression Model" 
        x.dist = "t"
        x.link = np.array
        temp = DynReg(formula=formula,data=data)
        temp.fit()

        for i in range(x.z_no-2):
            x.latent_variables.z_list[i].start = temp.latent_variables.get_z_values()[i+1]

        def temp_function(params):
            return -np.sum(ss.t.logpdf(x=x.data,df=np.exp(params[0]),
                loc=np.ones(x.data.shape[0])*params[1], scale=np.exp(params[2])))

        p = optimize.minimize(temp_function,np.array([2.0,0.0,-1.0]),method='L-BFGS-B')

        x.latent_variables.z_list[-1].start = p.x[0]
        x.latent_variables.z_list[-2].start = p.x[2]

        def draw_variable(loc,scale,shape,skewness,nsims):
            return loc + scale*np.random.standard_t(shape,nsims)

        x.draw_variable = draw_variable
        x.m_likelihood_markov_blanket = x.t_likelihood_markov_blanket

        return x

    @classmethod
    def skewt(cls,formula,data):
        """ Creates skewt-distributed state space model

        Parameters
        ----------
        data : np.array
            Contains the time series

        integ : int (default : 0)
            Specifies how many time to difference the time series.

        target : str (pd.DataFrame) or int (np.array)
            Specifies which column name or array index to use. By default, first
            column/array will be selected as the dependent variable.

        Returns
        ----------
        - NLLEV.skewt object
        """     

        x = NDynReg(formula=formula,data=data)
        
        x.latent_variables.add_z('Skewness', fam.Flat(transform='exp'), fam.Normal(0,3))        
        x.latent_variables.add_z('Signal^2 irregular', fam.Flat(transform='exp'), fam.Normal(0,3))
        x.latent_variables.add_z('v', fam.Flat(transform='exp'), fam.Normal(0,3))
        x.z_no += 3

        x._approximating_model = x._skewt_approximating_model
        x.meas_likelihood = x.skewt_likelihood
        x.model_name = "skewt-distributed Dynamic Regression Model" 
        x.dist = "skewt"
        x.link = np.array
        temp = DynReg(formula=formula,data=data)
        temp.fit()

        for i in range(x.z_no-3):
            x.latent_variables.z_list[i].start = temp.latent_variables.get_z_values()[i+1]

        def temp_function(params):
            return -np.sum(fam.Skewt.logpdf_internal(x=x.data,df=np.exp(params[0]),
                loc=np.ones(x.data.shape[0])*params[1], scale=np.exp(params[2]),gamma=np.exp(params[3])))

        p = optimize.minimize(temp_function,np.array([2.0,0.0,-1.0,1.0]),method='L-BFGS-B')

        x.latent_variables.z_list[-1].start = p.x[0]
        x.latent_variables.z_list[-2].start = p.x[2]
        x.latent_variables.z_list[-3].start = p.x[3]

        def draw_variable(loc,scale,shape,skewness,nsims):
            return loc + scale*fam.Skewt.rvs(shape,skewness,nsims)

        x.draw_variable = draw_variable
        x.m_likelihood_markov_blanket = x.skewt_likelihood_markov_blanket

        return x


    def neg_logposterior(self,beta):
        """ Returns negative log posterior

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            State matrix

        Returns
        ----------
        Negative log posterior
        """
        post = self.neg_loglik(beta)
        for k in range(0,self.z_no):
            post += -self.latent_variables.z_list[k].prior.logpdf(beta[k])
        return post     

    def state_likelihood_markov_blanket(self,beta,alpha,col_no):
        """ Returns Markov blanket of the states given the evolution latent variables

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            State matrix

        Returns
        ----------
        State likelihood
        """     
        _, _, _, Q = self._ss_matrices(beta)
        state_terms = np.append(0,ss.norm.logpdf(alpha[col_no][1:]-alpha[col_no][:-1],loc=0,scale=np.power(Q[col_no][col_no],0.5)))
        blanket = state_terms
        blanket[:-1] = blanket[:-1] + blanket[1:]
        return blanket

    def state_likelihood(self,beta,alpha):
        """ Returns likelihood of the states given the evolution latent variables

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            State matrix

        Returns
        ----------
        State likelihood
        """

        _, _, _, Q = self._ss_matrices(beta)
        state_lik = 0
        for i in range(alpha.shape[0]):
            state_lik += np.sum(ss.norm.logpdf(alpha[i][1:]-alpha[i][:-1],loc=0,scale=np.power(Q[i][i],0.5))) 
        return state_lik

    def loglik(self,beta,alpha):
        """ Creates negative loglikelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        Negative loglikelihood
        """     

        return (self.state_likelihood(beta,alpha) + self.meas_likelihood(beta,alpha))

    def neg_loglik(self,beta):
        """ Creates negative loglikelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        Negative loglikelihood
        """     
        states = np.zeros([self.state_no, self.data.shape[0]])
        for state_i in range(self.state_no):
            states[state_i,:] = beta[(self.z_no + (self.data.shape[0]*state_i)):(self.z_no + (self.data.shape[0]*(state_i+1)))]
        return -self.loglik(beta[:self.z_no],states) 

    def fit(self,optimizer='RMSProp',iterations=3000,print_progress=True,start_diffuse=False):
        """ Fits the model

        Parameters
        ----------
        optimizer : string
            Stochastic optimizer: either RMSProp or ADAM.

        iterations: int
            How many iterations to run

        print_progress : bool
            Whether tp print the ELBO progress or not
        
        start_diffuse : bool
            Whether to start from diffuse values (if not: use approx Gaussian)
        
        Returns
        ----------
        BBVI fit object
        """             
        return self._bbvi_fit(self.neg_logposterior,optimizer=optimizer,print_progress=print_progress,start_diffuse=start_diffuse,iterations=iterations)

    def _bbvi_fit(self,posterior,optimizer='RMSProp',iterations=3000,print_progress=True,start_diffuse=False,**kwargs):
        """ Performs Black Box Variational Inference

        Parameters
        ----------
        posterior : method
            Hands bbvi_fit a posterior object

        optimizer : string
            Stochastic optimizer: either RMSProp or ADAM.

        iterations: int
            How many iterations for BBVI

        Returns
        ----------
        BBVIResults object
        """

        # Starting parameters
        phi = self.latent_variables.get_z_starting_values()

        # Starting values for approximate distribution
        for i in range(len(self.latent_variables.z_list)):
            approx_dist = self.latent_variables.z_list[i].q
            if isinstance(approx_dist, fam.Normal):
                self.latent_variables.z_list[i].q.loc = phi[i]
                self.latent_variables.z_list[i].q.scale = -3.0

        q_list = [k.q for k in self.latent_variables.z_list]

        # Get starting values for states
        T, Z, R, Q = self._ss_matrices(phi)

        H, mu = self._approximating_model(phi,T,Z,R,Q)

        a, V = self.smoothed_state(self.data,phi,H,mu)

        for state in range(self.state_no):
            V[0][0][0] = V[0][0][-1] 
            for item in range(self.data.shape[0]):
                if start_diffuse is False:
                    q_list.append(fam.Normal(a[state][item],np.log(np.sqrt(np.abs(V[0][state][item])))))
                else:
                    q_list.append(fam.Normal(0,-3))

        bbvi_obj = ifr.CBBVI(posterior,self.log_p_blanket,q_list,24,optimizer,iterations)

        if print_progress is False:
            bbvi_obj.printer = False
        q, q_params, q_ses = bbvi_obj.run()

        self.latent_variables.set_z_values(q_params[:self.z_no],'BBVI',np.exp(q_ses[:self.z_no]),None)    

        for k in range(len(self.latent_variables.z_list)):
            self.latent_variables.z_list[k].q = q[k]

        # Theta values and states
        states = q_params[self.z_no:self.z_no+self.data.shape[0]]
        states_var = np.exp(q_ses[self.z_no:self.z_no+self.data.shape[0]])

        for state_i in range(1,self.state_no):
            states = np.vstack((states,q_params[(self.z_no+(self.data.shape[0]*state_i)):(self.z_no+(self.data.shape[0]*(state_i+1)))]))
            states_var = np.vstack((states_var,np.exp(q_ses[(self.z_no+(self.data.shape[0]*state_i)):(self.z_no+(self.data.shape[0]*(state_i+1)))])))

        if self.state_no == 1:
            states = np.array([states])
            states_var = np.array([states_var])

        theta = np.sum(self.X*states.T,axis=1)          
        Y = self.data
        scores = None
        X_names = self.X_names
        self.states = states
        self.states_var = states_var

        return res.BBVISSResults(data_name=self.data_name,X_names=X_names,model_name=self.model_name,
            model_type=self.model_type, latent_variables=self.latent_variables,data=Y,index=self.index,
            multivariate_model=self.multivariate_model,objective=posterior(q_params), 
            method='BBVI',ses=q_ses[:self.z_no],signal=theta,scores=scores,
            z_hide=self._z_hide,max_lag=self.max_lag,states=states,states_var=states_var)

    def exponential_likelihood(self,beta,alpha):
        """ Creates Exponential loglikelihood of the data given the states

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        Exponential loglikelihood
        """     
        return np.sum(ss.expon.logpdf(self.data,1/np.exp(np.sum(self.X*alpha.T,axis=1))))

    def exponential_likelihood_markov_blanket(self,beta,alpha):
        """ Creates Expnonential Markov blanket for each state

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        Exponential loglikelihood
        """     
        return ss.expon.logpdf(self.data,1/np.exp(np.sum(self.X*alpha.T,axis=1)))

    def laplace_likelihood(self,beta,alpha):
        """ Creates Poisson loglikelihood of the data given the states

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        Laplace loglikelihood
        """     
        return np.sum(ss.laplace.logpdf(self.data,np.sum(self.X*alpha.T,axis=1),scale=self.latent_variables.z_list[-1].prior.transform(beta[-1])))

    def laplace_likelihood_markov_blanket(self,beta,alpha):
        """ Creates Laplace Markov blanket for each state

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent_variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        Laplace Markov Blanket
        """     
        return ss.laplace.logpdf(self.data,np.sum(self.X*alpha.T,axis=1),scale=self.latent_variables.z_list[-1].prior.transform(beta[-1]))

    def poisson_likelihood(self,beta,alpha):
        """ Creates Poisson loglikelihood of the data given the states

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        Poisson loglikelihood
        """     
        return np.sum(ss.poisson.logpmf(self.data,np.exp(np.sum(self.X*alpha.T,axis=1))))

    def poisson_likelihood_markov_blanket(self,beta,alpha):
        """ Creates Poisson Markov blanket for each state

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        Poisson Markov Blanket
        """     
        return ss.poisson.logpmf(self.data,np.exp(np.sum(self.X*alpha.T,axis=1)))

    def t_likelihood(self,beta,alpha):
        """ Creates t loglikelihood of the date given the states

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        t loglikelihood
        """     
        return np.sum(ss.t.logpdf(x=self.data,
            df=self.latent_variables.z_list[-1].prior.transform(beta[-1]),
            loc=np.sum(self.X*alpha.T,axis=1),
            scale=self.latent_variables.z_list[-2].prior.transform(beta[-2])))

    def t_likelihood_markov_blanket(self,beta,alpha):
        """ Creates t Markov blanket for each state

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        t Markov Blanket
        """     
        return ss.t.logpdf(x=self.data,
            df=self.latent_variables.z_list[-1].prior.transform(beta[-1]),
            loc=np.sum(self.X*alpha.T,axis=1),
            scale=self.latent_variables.z_list[-2].prior.transform(beta[-2]))

    def skewt_likelihood(self,beta,alpha):
        """ Creates skewt loglikelihood of the date given the states

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        skewt loglikelihood
        """     
        return np.sum(fam.Skewt.logpdf_internal(x=self.data,
            df=self.latent_variables.z_list[-1].prior.transform(beta[-1]),
            loc=np.sum(self.X*alpha.T,axis=1),
            scale=self.latent_variables.z_list[-2].prior.transform(beta[-2]),  gamma=self.latent_variables.z_list[-3].prior.transform(beta[-3])))

    def skewt_likelihood_markov_blanket(self,beta,alpha):
        """ Creates skewt Markov blanket for each state

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        skewt Markov Blanket
        """     
        return fam.Skewt.logpdf_internal(x=self.data,
            df=self.latent_variables.z_list[-1].prior.transform(beta[-1]),
            loc=np.sum(self.X*alpha.T,axis=1),
            scale=self.latent_variables.z_list[-2].prior.transform(beta[-2]),  gamma=self.latent_variables.z_list[-3].prior.transform(beta[-3]))

    def markov_blanket(self,beta,alpha):
        """ Creates total Markov blanket for states

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        Markov blanket for states
        """ 
        likelihood_blanket = self.m_likelihood_markov_blanket(beta,alpha)
        state_blanket = self.state_likelihood_markov_blanket(beta,alpha,0)
        for i in range(self.state_no-1):
            likelihood_blanket = np.append(likelihood_blanket,self.m_likelihood_markov_blanket(beta,alpha))
            state_blanket = np.append(state_blanket,self.state_likelihood_markov_blanket(beta,alpha,i+1))
        return likelihood_blanket + state_blanket
        
    def evo_blanket(self,beta,alpha):
        """ Creates Markov blanket for the evolution latent variables

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        alpha : np.array
            A vector of states

        Returns
        ----------
        Markov blanket for evolution latent variables
        """ 
        evo_blanket = np.zeros(self.state_no)
        for i in range(evo_blanket.shape[0]):
            evo_blanket[i] = self.state_likelihood_markov_blanket(beta,alpha,i).sum()

        if self.dist in ['t']:
            evo_blanket = np.append([self.m_likelihood_markov_blanket(beta,alpha).sum()]*2,evo_blanket)
        if self.dist in ['skewt']:
            evo_blanket = np.append([self.m_likelihood_markov_blanket(beta,alpha).sum()]*3,evo_blanket)            
        elif self.dist in ['Laplace']:
            evo_blanket = np.append([self.m_likelihood_markov_blanket(beta,alpha).sum()],evo_blanket)

        return evo_blanket

    def log_p_blanket(self,beta):
        """ Creates complete Markov blanket for latent variables

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        Markov blanket for latent variables
        """     
        states = np.zeros([self.state_no, self.data.shape[0]])
        for state_i in range(self.state_no):
            states[state_i,:] = beta[(self.z_no + (self.data.shape[0]*state_i)):(self.z_no + (self.data.shape[0]*(state_i+1)))]     
        
        return np.append(self.evo_blanket(beta,states),self.markov_blanket(beta,states))

    def plot_predict(self,h=5,past_values=20,intervals=True,oos_data=None,**kwargs):        
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show 95% prediction intervals for the forecast?

        oos_data : pd.DataFrame
            OOS data to use; needs to be same format (columns) as original data

        Returns
        ----------
        - Plot of the forecast
        """     

        figsize = kwargs.get('figsize',(10,7))

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            # Retrieve data, dates and (transformed) latent variables
            scale, shape, skewness = self._get_scale_and_shape()

            # Retrieve data, dates and (transformed) latent variables
            date_index = self.shift_dates(h)
            simulations = 10000
            sim_vector = np.zeros([simulations,h])

            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            full_X = self.X.copy()
            full_X = np.append(full_X,X_oos,axis=0)
            Z = full_X
            a = self.states

            # Retrieve data, dates and (transformed) latent variables         
            smoothed_series = np.zeros(h)
            for t in range(h):
                smoothed_series[t] = self.link(np.dot(Z[self.y.shape[0]+t],a[:,-1]))

            for n in range(0,simulations):  
                rnd_q = np.zeros((self.state_no,h))
                coeff_sim = np.zeros((self.state_no,h))

                # TO DO: vectorize this (easy)
                for state in range(self.state_no):
                    rnd_q[state] = np.random.normal(0,np.sqrt(self.latent_variables.get_z_values(transformed=True)[state]),h)

                for t in range(0,h):
                    if t == 0:
                        for state in range(self.state_no):
                            coeff_sim[state][t] = a[state][-1] + rnd_q[state][t]
                    else:
                        for state in range(self.state_no):
                            coeff_sim[state][t] = coeff_sim[state][t-1] + rnd_q[state][t]

                sim_vector[n] = self.draw_variable(loc=self.link(np.sum(coeff_sim.T*Z[self.y.shape[0]:self.y.shape[0]+h,:],axis=1)),shape=shape,scale=scale,skewness=skewness,nsims=h)

            sim_vector = np.transpose(sim_vector)
            forecasted_values = smoothed_series
            previous_value = self.data[-1]

            plt.figure(figsize=figsize) 

            if intervals == True:
                plt.fill_between(date_index[-h-1:], np.insert([np.percentile(i,5) for i in sim_vector],0,previous_value), 
                    np.insert([np.percentile(i,95) for i in sim_vector],0,previous_value), alpha=0.2,label="95 C.I.")   

            plot_values = np.append(self.data[-past_values:],forecasted_values)
            plot_index = date_index[-h-past_values:]

            plt.plot(plot_index,plot_values,label=self.data_name)
            plt.title("Forecast for " + self.data_name)
            plt.xlabel("Time")
            plt.ylabel(self.data_name)
            plt.show()

    def plot_fit(self,intervals=True,**kwargs):
        """ Plots the fit of the model

        Parameters
        ----------
        intervals : Boolean
            Whether to plot 95% confidence interval of states

        Returns
        ----------
        None (plots data and the fit)
        """

        figsize = kwargs.get('figsize',(10,7))

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            date_index = copy.deepcopy(self.index)
            date_index = date_index[:self.y.shape[0]+1]

            mu = self.states
            var = self.states_var
            # Create smoothed/filtered aggregate series
            _, Z, _, _ = self._ss_matrices(self.latent_variables.get_z_values())
            smoothed_series = np.zeros(self.y.shape[0])

            for t in range(0,self.y.shape[0]):
                smoothed_series[t] = np.dot(Z[t],mu[:,t])

            plt.figure(figsize=figsize) 
            
            plt.subplot(self.state_no+1, 1, 1)
            plt.title(self.y_name + " Raw and Smoothed")    
            plt.plot(date_index,self.data,label='Data')
            plt.plot(date_index,self.link(smoothed_series),label='Smoothed Series',c='black')
            plt.legend(loc=2)
            
            for coef in range(0,self.state_no):
                V_coef = self.states_var[coef]
                plt.subplot(self.state_no+1, 1, 2+coef)
                plt.title("Beta " + self.X_names[coef]) 
                states_upper_95 = self.states[coef] + 1.98*np.sqrt(V_coef)
                states_lower_95 = self.states[coef] - 1.98*np.sqrt(V_coef)

                if intervals == True:
                    alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                    plt.fill_between(date_index, states_lower_95, states_upper_95, alpha=0.15,label='95% C.I.') 

                plt.plot(date_index,mu[coef,:],label='Coefficient')
                plt.legend(loc=2)               
            
            plt.show()  

    def predict(self,h=5,oos_data=None):        
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        oos_data : pd.DataFrame
            Data for the variables to be used out of sample (ys can be NaNs)

        Returns
        ----------
        - pd.DataFrame with predictions
        """     

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            # Sort/manipulate the out-of-sample data
            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            full_X = self.X.copy()
            full_X = np.append(full_X,X_oos,axis=0)
            Z = full_X
            a = self.states

            # Retrieve data, dates and (transformed) latent variables         
            smoothed_series = np.zeros(h)
            for t in range(h):
                smoothed_series[t] = self.link(np.dot(Z[self.y.shape[0]+t],a[:,-1]))

            date_index = self.shift_dates(h)

            result = pd.DataFrame(smoothed_series)
            result.rename(columns={0:self.y_name}, inplace=True)
            result.index = date_index[-h:]

            return result

    def predict_is(self,h=5):
        """ Makes dynamic in-sample predictions with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        predictions = []

        for t in range(0,h):
            data1 = self.data_original.iloc[0:-h+t,:]
            data2 = self.data_original.iloc[-h+t:,:]            
            if self.dist == 'Poisson':
                x = NDynReg.Poisson(formula=self.formula,data=data1)
            elif self.dist == 't':
                x = NDynReg.t(formula=self.formula,data=data1)
            elif self.dist == 'Laplace':
                x = NDynReg.Laplace(formula=self.formula,data=data1)
            elif self.dist == 'Exponential':
                x = NDynReg.Exponential(formula=self.formula,data=data1)                            
            x.fit(print_progress=False)
            if t == 0:
                predictions = x.predict(1,oos_data=data2)
            else:
                predictions = pd.concat([predictions,x.predict(1,oos_data=data2)])
        
        predictions.rename(columns={0:self.data_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions

    def plot_predict_is(self,h=5,**kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        Returns
        ----------
        - Plot of the forecast against data 
        """     

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        date_index = self.index[-h:]
        predictions = self.predict_is(h)
        data = self.data[-h:]

        plt.plot(date_index,data,label='Data')
        plt.plot(date_index,predictions,label='Predictions',c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def simulation_smoother(self,beta):
        """ Durbin and Koopman simulation smoother - simulates from states 
        given model latent variables and observations

        Parameters
        ----------

        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        - A simulated state evolution
        """         

        T, Z, R, Q = self._ss_matrices(beta)
        H, mu = self._approximating_model(beta,T,Z,R,Q)

        # Generate e_t+ and n_t+
        rnd_h = np.random.normal(0,np.sqrt(H),self.data.shape[0])
        q_dist = ss.multivariate_normal([0.0,0.0], Q,allow_singular=True)
        rnd_q = q_dist.rvs(self.data.shape[0])

        # Generate a_t+ and y_t+
        a_plus = np.zeros((T.shape[0],self.data.shape[0])) 
        y_plus = np.zeros(self.data.shape[0])

        for t in range(0,self.data.shape[0]):
            if t == 0:
                a_plus[:,t] = np.dot(T,a_plus[:,t]) + rnd_q[t]
                y_plus[t] = mu[t] + np.dot(Z,a_plus[:,t]) + rnd_h[t]
            else:
                if t != self.data.shape[0]:
                    a_plus[:,t] = np.dot(T,a_plus[:,t-1]) + rnd_q[t]
                    y_plus[t] = mu[t] + np.dot(Z,a_plus[:,t]) + rnd_h[t]

        alpha_hat, _ = self.smoothed_state(self.data,beta, H, mu)
        alpha_hat_plus, _ = self.smoothed_state(y_plus,beta, H, mu)
        alpha_tilde = alpha_hat - alpha_hat_plus + a_plus
        
        return alpha_tilde

    def smoothed_state(self,data,beta,H,mu):
        """ Creates smoothed state estimate given state matrices and 
        parameters.

        Parameters
        ----------

        data : np.array
            Data to be smoothed

        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        - Smoothed states
        """         

        T, Z, R, Q = self._ss_matrices(beta)
        alpha, V = nld_univariate_KFS(data,Z,H,T,Q,R,mu)
        return alpha, V


import numpy as np
import scipy.stats as ss
import scipy.special as sp

from .family import Family
from .flat import Flat
from .normal import Normal

from .gas_recursions import gas_recursion_t_orderone, gas_recursion_t_ordertwo
from .gas_recursions import gasx_recursion_t_orderone, gasx_recursion_t_ordertwo
from .gas_recursions import gas_llev_recursion_t_orderone, gas_llev_recursion_t_ordertwo
from .gas_recursions import gas_llt_recursion_t_orderone, gas_llt_recursion_t_ordertwo
from .gas_recursions import gas_reg_recursion_t_orderone, gas_reg_recursion_t_ordertwo


class t(Family):
    """ 
    Student t Distribution
    ----
    This class contains methods relating to the Student t distribution for time series.
    """

    def __init__(self, loc=0.0, scale=1.0, df=8.0, transform=None, **kwargs):
        """
        Parameters
        ----------
        loc : float
            Location parameter for the t distribution

        scale : float
            Scale parameter for the t distribution

        df : float
            Degrees of freedom parameter for the t distribution

        transform : str
            Whether to apply a transformation to the location variable - e.g. 'exp' or 'logit'
        """
        super(t, self).__init__(transform)
        self.loc0 = loc
        self.scale0 = scale
        self.df0 = df
        self.covariance_prior = False

        self.gradient_only = kwargs.get('gradient_only', False) # used for GAS t models
        if self.gradient_only is True:
            self.score_function = self.first_order_score
        else:
            self.score_function = self.second_order_score

    def approximating_model(self, beta, T, Z, R, Q, h_approx, data):
        """ Creates approximating Gaussian state space model for t measurement density
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm
        
        h_approx : float
            The variance of the measurement density
        
        data: np.array
            The univariate time series data

        Returns
        ----------
        H : np.array
            Approximating measurement variance matrix
        
        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(data.shape[0])*h_approx
        mu = np.zeros(data.shape[0])

        return H, mu

    def approximating_model_reg(self, beta, T, Z, R, Q, h_approx, data, X, state_no):
        """ Creates approximating Gaussian state space model for t measurement density
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm
        
        h_approx : float
            The variance of the measurement density
        
        data: np.array
            The univariate time series data

        X: np.array
            The regressors

        state_no : int
            Number of states

        Returns
        ----------
        H : np.array
            Approximating measurement variance matrix
        
        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(data.shape[0])*h_approx
        mu = np.zeros(data.shape[0])

        return H, mu

    @staticmethod
    def build_latent_variables():
        """ Builds additional latent variables for this family

        Returns
        ----------
        - A list of lists (each sub-list contains latent variable information)
        """
        lvs_to_build = []
        lvs_to_build.append(['t Scale', Flat(transform='exp'), Normal(0, 3), 0.01])
        lvs_to_build.append(['v', Flat(transform='exp'), Normal(0, 3), 2.5])
        return lvs_to_build

    @staticmethod
    def draw_variable(loc, scale, shape, skewness, nsims):
        """ Draws random variables from t distribution

        Parameters
        ----------
        loc : float
            location parameter for the distribution

        scale : float
            scale parameter for the distribution

        shape : float
            tail thickness parameter for the distribution

        skewness : float
            skewness parameter for the distribution

        nsims : int or list
            number of draws to take from the distribution

        Returns
        ----------
        - Random draws from the distribution
        """
        return loc + scale*np.random.standard_t(shape,nsims)

    @staticmethod
    def first_order_score(y, mean, scale, shape, skewness):
        """ GAS t Update term using gradient only - native Python function

        Parameters
        ----------
        y : float
            datapoint for the time series

        mean : float
            location parameter for the t distribution

        scale : float
            scale parameter for the t distribution

        shape : float
            tail thickness parameter for the t distribution

        skewness : float
            skewness parameter for the t distribution

        Returns
        ----------
        - Score of the t family
        """
        return ((shape+1)/shape)*(y-mean)/(np.power(scale,2) + (np.power(y-mean,2)/shape))

    def logpdf(self, mu):
        """
        Log PDF for t prior

        Parameters
        ----------
        mu : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - log(p(mu))
        """
        if self.transform is not None:
            mu = self.transform(mu)    
        return ss.t.logpdf(mu, df=self.df0, loc=self.loc0, scale=self.scale0) 

    @staticmethod
    def markov_blanket(y, mean, scale, shape, skewness):
        """ Markov blanket for the Poisson distribution

        Parameters
        ----------
        y : np.ndarray
            univariate time series

        mean : np.ndarray
            array of location parameters for the t distribution

        scale : float
            scale parameter for the t distribution

        shape : float
            tail thickness parameter for the t distribution

        skewness : float
            skewness parameter for the t distribution

        Returns
        ----------
        - Markov blanket of the t family
        """
        return ss.t.logpdf(x=y, df=shape, loc=mean, scale=scale)

    @staticmethod
    def setup():
        """ Returns the attributes of this family

        Notes
        ----------
        - scale notes whether family has a variance parameter (sigma)
        - shape notes whether family has a tail thickness parameter (nu)
        - skewness notes whether family has a skewness parameter (gamma)
        - mean_transform is a function which transforms the location parameter
        - cythonized notes whether the family has cythonized routines
        
        Returns
        ----------
        - model name, link function, scale, shape, skewness, mean_transform, cythonized
        """
        name = "t"
        link = np.array
        scale = True
        shape = True
        skewness = False
        mean_transform = np.array
        cythonized = True
        return name, link, scale, shape, skewness, mean_transform, cythonized

    @staticmethod
    def neg_loglikelihood(y, mean, scale, shape, skewness):
        """ Negative loglikelihood function

        Parameters
        ----------
        y : np.ndarray
            univariate time series

        mean : np.ndarray
            array of location parameters for the t distribution

        scale : float
            scale parameter for the t distribution

        shape : float
            tail thickness parameter for the t distribution

        skewness : float
            skewness parameter for the t distribution

        Returns
        ----------
        - Negative loglikelihood of the t family
        """
        return -np.sum(ss.t.logpdf(x=y, df=shape, loc=mean, scale=scale))

    def pdf(self, mu):
        """
        PDF for t prior

        Parameters
        ----------
        mu : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - p(mu)
        """
        if self.transform is not None:
            mu = self.transform(mu)    
        return ss.t.pdf(mu, df=self.df0, loc=self.loc0, scale=self.scale0) 

    @staticmethod
    def reg_score_function(X, y, mean, scale, shape, skewness):
        """ GAS t Regression Update term using gradient only - native Python function

        Parameters
        ----------
        X : float
            datapoint for the right hand side variable
    
        y : float
            datapoint for the time series

        mean : float
            location parameter for the t distribution

        scale : float
            scale parameter for the t distribution

        shape : float
            tail thickness parameter for the t distribution

        skewness : float
            skewness parameter for the t distribution

        Returns
        ----------
        - Score of the t family
        """
        return ((shape+1)/shape)*((y-mean)*X)/(np.power(scale,2)+np.power((y-mean),2)/shape)

    @staticmethod
    def second_order_score(y, mean, scale, shape, skewness):
        """ GAS t Update term potentially using second-order information - native Python function

        Parameters
        ----------
        y : float
            datapoint for the time series

        mean : float
            location parameter for the t distribution

        scale : float
            scale parameter for the t distribution

        shape : float
            tail thickness parameter for the t distribution

        skewness : float
            skewness parameter for the t distribution

        Returns
        ----------
        - Adjusted score of the t family
        """
        return ((shape+1)/shape)*(y-mean)/(np.power(scale,2) + (np.power(y-mean,2)/shape))/((shape+1)*((np.power(scale,2)*shape) - np.power(y-mean,2))/np.power((np.power(scale,2)*shape) + np.power(y-mean,2),2))

    # Optional Cythonized recursions below for GAS t models

    @staticmethod
    def gradient_recursion():
        """ GAS t Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS t model - gradient only
        """
        return gas_recursion_t_orderone

    @staticmethod
    def newton_recursion():
        """ GAS t Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS t model - adjusted score
        """
        return gas_recursion_t_ordertwo

    @staticmethod
    def gradientx_recursion():
        """ GASX t Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GASX t model - gradient only
        """
        return gasx_recursion_t_orderone

    @staticmethod
    def newtonx_recursion():
        """ GASX t Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GASX t model - adjusted score
        """
        return gasx_recursion_t_ordertwo

    @staticmethod
    def gradientllev_recursion():
        """ GAS Local Level t Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Local Level t model - gradient only
        """
        return gas_llev_recursion_t_orderone

    @staticmethod
    def newtonllev_recursion():
        """ GAS Local Level t Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Local Level t model - adjusted score
        """
        return gas_llev_recursion_t_ordertwo

    @staticmethod
    def gradientllt_recursion():
        """ GAS Local Linear Trend t Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Local Linear Trend t model - gradient only
        """
        return gas_llt_recursion_t_orderone

    @staticmethod
    def newtonllt_recursion():
        """ GAS Local Linear Trend t Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Local Linear Trend t model - adjusted score
        """
        return gas_llt_recursion_t_ordertwo

    @staticmethod
    def gradientreg_recursion():
        """ GAS Dynamic Regression t Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Dynamic Regression t model - gradient only
        """
        return gas_reg_recursion_t_orderone

    @staticmethod
    def newtonreg_recursion():
        """ GAS Dynamic Regression t Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Dynamic Regression t model - adjusted score
        """
        return gas_reg_recursion_t_ordertwo


from numpy import abs, exp, power, array, sqrt, pi
from scipy.special import gamma

#TODO: This file should eventually be replaced, by moving the existing functions (used for GARCH based models)
# into the GARCH folder, with Cythonizations

class Score(object):

    @staticmethod
    def score(y,loc,scale,shape):
        pass

    @staticmethod
    def adj_score(y,loc,scale,shape):
        pass

class BetatScore(Score):

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def mu_score(y,loc,scale,shape):
        try:
            return (((shape+1.0)*power(y-loc,2))/float(shape*exp(scale) + power(y-loc,2))) - 1.0
        except:
            return -1.0

    @staticmethod
    def mu_adj_score(y,loc,scale,shape):
        try:
            return (((shape+1.0)*power(y-loc,2))/float(shape*exp(scale) + power(y-loc,2))) - 1.0
        except:
            return -1.0

    def score(self,y,loc,scale,shape):
        return array([self.mu_score(y,loc,scale,shape)])

    def adj_score(self,y,loc,scale,shape):
        return array([self.mu_adj_score(y,loc,scale,shape)])

class SkewBetatScore(Score):

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def tv_variate_exp(df):
        return (sqrt(df)*gamma((df-1.0)/2.0))/(sqrt(pi)*gamma(df/2.0))

    @staticmethod
    def mu_score(y,loc,scale,shape,skewness):
        try:
            if (y-loc)>=0:
                return (((shape+1.0)*power(y-loc,2))/float(power(skewness,2)*shape*exp(scale) + power(y-loc,2))) - 1.0
            else:
                return (((shape+1.0)*power(y-loc,2))/float(power(skewness,-2)*shape*exp(scale) + power(y-loc,2))) - 1.0    
        except:
            return -1.0

    @staticmethod
    def mu_adj_score(y,loc,scale,shape,skewness):
        try:
            if (y-loc)>=0:
                return (((shape+1.0)*power(y-loc,2))/float(power(skewness,2)*shape*exp(scale) + power(y-loc,2))) - 1.0
            else:
                return (((shape+1.0)*power(y-loc,2))/float(power(skewness,-2)*shape*exp(scale) + power(y-loc,2))) - 1.0    
        except:
            return -1.0

    def score(self,y,loc,scale,shape,skewness):
        return array([self.mu_score(y,loc,scale,shape,skewness)])

    def adj_score(self,y,loc,scale,shape,skewness):
        return array([self.mu_adj_score(y,loc,scale,shape,skewness)])


import numpy as np
import scipy.stats as ss
import scipy.special as sp

from .. import inference as ifr

from .family import Family

from scipy.stats import invwishart


class InverseWishart(Family):
    """ 
    Inverse Wishart Distribution
    ----
    This class contains methods relating to the inverse wishart distribution for time series.
    """

    def __init__(self, v, Psi, transform=None, **kwargs):
        """
        Parameters
        ----------
        v : float
            Nu parameter for the Inverse Wishart distribution

        Psi : float
            Psi parameter for the Inverse Wishart distribution

        transform : str
            Whether to apply a transformation - e.g. 'exp' or 'logit'
        """
        super(InverseWishart, self).__init__(transform)

        self.covariance_prior = True
        self.v = v
        self.Psi = Psi

    def logpdf(self, X):
        """
        Log PDF for Inverse Wishart prior

        Parameters
        ----------
        X : float
            Covariance matrix for which the prior is being formed over

        Returns
        ----------
        - log(p(X))
        """
        return invwishart.logpdf(X, df=self.v, scale=self.Psi)

    def pdf(self, X):
        """
        PDF for Inverse Wishart prior

        Parameters
        ----------
        x : float
            Covariance matrix for which the prior is being formed over

        Returns
        ----------
        - p(x)
        """              
        return invwishart.pdf(X, df=self.v, scale=self.Psi)


import copy
import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
from patsy import dmatrices, dmatrix, demo_data

from .. import families as fam
from .. import tsm as tsm
from .. import data_check as dc

from .kalman import *

class DynReg(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** DYNAMIC REGRESSION MODEL ****

    Parameters
    ----------

    formula : string
        patsy string describing the regression

    data : pd.DataFrame
        Field to specify the data that will be used
    """

    def __init__(self,formula,data):

        # Initialize TSM object
        super(DynReg,self).__init__('DynReg')

        # Latent variables
        self.max_lag = 0
        self._z_hide = 0 # Whether to cutoff variance latent variables from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"
        self.model_name = "Dynamic Linear Regression"
        self.multivariate_model = False

        # Format the data
        self.is_pandas = True # This is compulsory for this model type
        self.data_original = data
        self.formula = formula
        self.y, self.X = dmatrices(formula, data)
        self.z_no = self.X.shape[1] + 1
        self.y_name = self.y.design_info.describe()
        self.data_name = self.y_name
        self.X_names = self.X.design_info.describe().split(" + ")
        self.y = np.array([self.y]).ravel()
        self.data = self.y
        self.X = np.array([self.X])[0]
        self.index = data.index

        self._create_latent_variables()

    def _create_latent_variables(self):
        """ Creates model latent variables

        Returns
        ----------
        None (changes model attributes)
        """

        self.latent_variables.add_z('Sigma^2 irregular', fam.Flat(transform='exp'), fam.Normal(0,3))

        for parm in range(self.z_no-1):
            self.latent_variables.add_z('Sigma^2 ' + self.X_names[parm], fam.Flat(transform='exp'), fam.Normal(0,3))

    def _forecast_model(self,beta,Z,h):
        """ Creates forecasted states and variances

        Parameters
        ----------
        beta : np.ndarray
            Contains untransformed starting values for latent variables

        Returns
        ----------
        a : np.ndarray
            Forecasted states

        P : np.ndarray
            Variance of forecasted states
        """     

        T, _, R, Q, H = self._ss_matrices(beta)
        return dl_univariate_kalman_fcst(self.data,Z,H,T,Q,R,0.0,h)

    def _model(self,data,beta):
        """ Creates the structure of the model

        Parameters
        ----------
        data : np.array
            Contains the time series

        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        a,P,K,F,v : np.array
            Filted states, filtered variances, Kalman gains, F matrix, residuals
        """     

        T, Z, R, Q, H = self._ss_matrices(beta)

        return dl_univariate_kalman(data,Z,H,T,Q,R,0.0)

    def _ss_matrices(self, beta):
        """ Creates the state space matrices required

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        T, Z, R, Q, H : np.array
            State space matrices used in KFS algorithm
        """     

        T = np.identity(self.z_no-1)
        H = np.identity(1)*self.latent_variables.z_list[0].prior.transform(beta[0])       
        Z = self.X
        R = np.identity(self.z_no-1)
        
        Q = np.identity(self.z_no-1)
        for i in range(0,self.z_no-1):
            Q[i][i] = self.latent_variables.z_list[i+1].prior.transform(beta[i+1])

        return T, Z, R, Q, H

    def neg_loglik(self,beta):
        """ Creates the negative log marginal likelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        The negative log logliklihood of the model
        """         
        _, _, _, F, v = self._model(self.y,beta)
        loglik = 0.0
        for i in range(0,self.y.shape[0]):
            loglik += np.linalg.slogdet(F[:,:,i])[1] + np.dot(v[i],np.dot(np.linalg.pinv(F[:,:,i]),v[i]))
        return -(-((self.y.shape[0]/2)*np.log(2*np.pi))-0.5*loglik.T[0].sum())

    def mb_neg_loglik(self, beta, mini_batch):
        """ Creates the negative log likelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        mini_batch : int
            Size of each mini batch of data

        Returns
        ----------
        The negative log logliklihood of the model
        """    

        rand_int =  np.random.randint(low=0, high=self.data.shape[0]-mini_batch-self.max_lag+1)
        sample = np.arange(start=rand_int, stop=rand_int+mini_batch)
        _, _, _, F, v = self._model(self.data[sample],beta)
        
        loglik = 0.0
        for i in range(0,len(sample)):
            loglik += np.linalg.slogdet(F[:,:,i])[1] + np.dot(v[i],np.dot(np.linalg.pinv(F[:,:,i]),v[i]))
        return -(-((len(sample)/2)*np.log(2*np.pi))-0.5*loglik.T[0].sum())

    def plot_predict(self, h=5, past_values=20, intervals=True, oos_data=None, **kwargs):        
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show 95% prediction intervals for the forecast?

        oos_data : pd.DataFrame
            Data for the variables to be used out of sample (ys can be NaNs)

        Returns
        ----------
        - Plot of the forecast
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))
        nsims = kwargs.get('nsims', 200)

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:

            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            full_X = self.X.copy()
            full_X = np.append(full_X,X_oos,axis=0)
            Z = full_X
            date_index = self.shift_dates(h)

            # Retrieve data, dates and (transformed) latent variables   
            if self.latent_variables.estimation_method in ['M-H']:
                lower_final = 0
                upper_final = 0
                plot_values_final = 0
                plot_index = date_index[-h-past_values:]

                for i in range(nsims):

                    t_params = self.draw_latent_variables(nsims=1).T[0]
                    a, P = self._forecast_model(t_params, Z, h)

                    smoothed_series = np.zeros(self.y.shape[0]+h)
                    series_variance = np.zeros(self.y.shape[0]+h)

                    for t in range(self.y.shape[0]+h):
                        smoothed_series[t] = np.dot(Z[t],a[:,t])
                        series_variance[t] = np.dot(np.dot(Z[t],P[:,:,t]),Z[t].T)

                    plot_values = smoothed_series[-h-past_values:]
                    lower = smoothed_series[-h:] - 1.96*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(t_params[0]),0.5)
                    upper = smoothed_series[-h:] + 1.96*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(t_params[0]),0.5)
                    lower_final += np.append(plot_values[-h-1], lower)
                    upper_final += np.append(plot_values[-h-1], upper)
                    plot_values_final += plot_values

                plot_values_final = plot_values_final / nsims
                lower_final = lower_final / nsims
                upper_final = upper_final / nsims

                plt.figure(figsize=figsize)
                if intervals == True:
                    plt.fill_between(date_index[-h-1:], lower_final, upper_final, alpha=0.2)            

                plt.plot(plot_index, plot_values_final)
                plt.title("Forecast for " + self.data_name)
                plt.xlabel("Time")
                plt.ylabel(self.data_name)
                plt.show()
            else:
                a, P = self._forecast_model(self.latent_variables.get_z_values(), h)
                plot_values = a[0][-h-past_values:]
                forecasted_values = a[0][-h:]

                smoothed_series = np.zeros(self.y.shape[0]+h)
                series_variance = np.zeros(self.y.shape[0]+h)

                for t in range(self.y.shape[0]+h):
                    smoothed_series[t] = np.dot(Z[t],a[:,t])
                    series_variance[t] = np.dot(np.dot(Z[t],P[:,:,t]),Z[t].T)

                lower = forecasted_values - 1.96*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(self.latent_variables.get_z_values()[0]),0.5)
                upper = forecasted_values + 1.96*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(self.latent_variables.get_z_values()[0]),0.5)
                lower = np.append(plot_values[-h-1],lower)
                upper = np.append(plot_values[-h-1],upper)

                plot_index = date_index[-h-past_values:]

                plt.figure(figsize=figsize)
                if intervals == True:
                    plt.fill_between(date_index[-h-1:], lower, upper, alpha=0.2)            

                plt.plot(plot_index,plot_values)
                plt.title("Forecast for " + self.data_name)
                plt.xlabel("Time")
                plt.ylabel(self.data_name)
                plt.show()

    def plot_fit(self,intervals=False,**kwargs):
        """ Plots the fit of the model

        Parameters
        ----------
        intervals : Boolean
            Whether to plot 95% confidence interval of states

        Returns
        ----------
        None (plots data and the fit)
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))
        series_type = kwargs.get('series_type','Smoothed')

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            date_index = copy.deepcopy(self.index)
            date_index = date_index[:self.y.shape[0]+1]

            if series_type == 'Smoothed':
                mu, V = self.smoothed_state(self.data,self.latent_variables.get_z_values())
            elif series_type == 'Filtered':
                mu, V, _, _, _ = self._model(self.data,self.latent_variables.get_z_values())
            else:
                mu, V = self.smoothed_state(self.data,self.latent_variables.get_z_values())

            # Create smoothed/filtered aggregate series
            _, Z, _, _, _ = self._ss_matrices(self.latent_variables.get_z_values())
            smoothed_series = np.zeros(self.y.shape[0])

            for t in range(0,self.y.shape[0]):
                smoothed_series[t] = np.dot(Z[t],mu[:,t])

            plt.figure(figsize=figsize) 
            
            plt.subplot(self.z_no+1, 1, 1)
            plt.title(self.y_name + " Raw and " + series_type)  
            plt.plot(date_index,self.data,label='Data')
            plt.plot(date_index,smoothed_series,label=series_type,c='black')
            plt.legend(loc=2)

            for coef in range(0,self.z_no-1):
                V_coef = V[0][coef][:-1]    
                plt.subplot(self.z_no+1, 1, 2+coef)
                plt.title("Beta " + self.X_names[coef]) 

                if intervals == True:
                    alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                    plt.fill_between(date_index[5:], mu[coef,0:mu.shape[1]-1][5:] + 1.98*np.sqrt(V_coef[5:]), mu[coef,0:mu.shape[1]-1][5:] - 1.98*np.sqrt(V_coef[5:]), alpha=0.15,label='95% C.I.') 
                plt.plot(date_index,mu[coef,0:mu.shape[1]-1],label='Data')
                plt.legend(loc=2)               
            
            plt.subplot(self.z_no+1, 1, self.z_no+1)
            plt.title("Measurement Error")
            plt.plot(date_index,self.data-smoothed_series,label='Irregular')
            plt.legend(loc=2)   

            plt.show()  

    def predict(self, h=5, intervals=False, oos_data=None, **kwargs):        
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        intervals : boolean (default: False)
            Whether to return prediction intervals

        oos_data : pd.DataFrame
            Data for the variables to be used out of sample (ys can be NaNs)

        Returns
        ----------
        - pd.DataFrame with predictions
        """     

        nsims = kwargs.get('nsims', 200)

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:

            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            full_X = self.X.copy()
            full_X = np.append(full_X,X_oos,axis=0)
            Z = full_X
            date_index = self.shift_dates(h)

            # Retrieve data, dates and (transformed) latent variables   
            if self.latent_variables.estimation_method in ['M-H']:
                lower_1_final = 0
                upper_99_final = 0
                lower_5_final = 0
                upper_95_final = 0
                forecasted_values_final = 0

                for i in range(nsims):
                    t_params = self.draw_latent_variables(nsims=1).T[0]
                    a, P = self._forecast_model(t_params, Z, h)

                    smoothed_series = np.zeros(h)
                    series_variance = np.zeros(h)

                    for t in range(h):
                        smoothed_series[t] = np.dot(Z[self.y.shape[0]+t],a[:,self.y.shape[0]+t])
                        series_variance[t] = np.dot(np.dot(Z[self.y.shape[0]+t],P[:,:,self.y.shape[0]+t]),Z[self.y.shape[0]+t].T)

                    forecasted_values = smoothed_series
                    lower_5 = smoothed_series - 1.96*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(t_params[0]),0.5)
                    upper_95 = smoothed_series + 1.96*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(t_params[0]),0.5)
                    lower_5_final += lower_5
                    upper_95_final += upper_95
                    lower_1 = smoothed_series - 2.575*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(t_params[0]),0.5)
                    upper_99 = smoothed_series + 2.575*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(t_params[0]),0.5)
                    lower_1_final += lower_1
                    upper_99_final += upper_99
                    forecasted_values_final += forecasted_values

                forecasted_values_final = forecasted_values_final / nsims
                lower_1_final = lower_1_final / nsims
                lower_5_final = lower_5_final / nsims
                upper_95_final = upper_95_final / nsims
                upper_99_final = upper_99_final / nsims

                if intervals is False:
                    result = pd.DataFrame(forecasted_values_final)
                    result.rename(columns={0:self.data_name}, inplace=True)
                else:
                    prediction_05 = lower_5_final
                    prediction_95 = upper_95_final
                    prediction_01 = lower_1_final
                    prediction_99 = upper_99_final

                    result = pd.DataFrame([forecasted_values_final, prediction_01, prediction_05, 
                        prediction_95, prediction_99]).T
                    result.rename(columns={0:self.data_name, 1: "1% Prediction Interval", 
                        2: "5% Prediction Interval", 3: "95% Prediction Interval", 4: "99% Prediction Interval"}, 
                        inplace=True)

                result.index = date_index[-h:]

                return result
     
            else:
                t_params = self.latent_variables.get_z_values()
                a, P = self._forecast_model(t_params, Z, h)
                smoothed_series = np.zeros(h)

                for t in range(h):
                    smoothed_series[t] = np.dot(Z[self.y.shape[0]+t],a[:,self.y.shape[0]+t])

                # Retrieve data, dates and (transformed) latent variables         
                forecasted_values = smoothed_series

                if intervals is False:
                    result = pd.DataFrame(forecasted_values)
                    result.rename(columns={0:self.data_name}, inplace=True)
                else:

                    series_variance = np.zeros(h)

                    for t in range(h):
                        series_variance[t] = np.dot(np.dot(Z[self.y.shape[0]+t],P[:,:,self.y.shape[0]+t]),Z[self.y.shape[0]+t].T)

                    prediction_05 = forecasted_values - 1.96*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(self.latent_variables.get_z_values()[0]),0.5)
                    prediction_95 = forecasted_values + 1.96*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(self.latent_variables.get_z_values()[0]),0.5)
                    prediction_01 = forecasted_values - 2.575*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(self.latent_variables.get_z_values()[0]),0.5)
                    prediction_99 = forecasted_values + 2.575*np.power(P[0][0][-h:] + self.latent_variables.z_list[0].prior.transform(self.latent_variables.get_z_values()[0]),0.5)

                    result = pd.DataFrame([forecasted_values, prediction_01, prediction_05, 
                        prediction_95, prediction_99]).T
                    result.rename(columns={0:self.data_name, 1: "1% Prediction Interval", 
                        2: "5% Prediction Interval", 3: "95% Prediction Interval", 4: "99% Prediction Interval"}, 
                        inplace=True)
     
                result.index = date_index[-h:]

                return result

    def predict_is(self, h=5, fit_once=True, fit_method='MLE', intervals=False):
        """ Makes dynamic in-sample predictions with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        fit_once : boolean
            (default: True) Fits only once before the in-sample prediction; if False, fits after every new datapoint

        fit_method : string
            Which method to fit the model with

        intervals : boolean
            Whether to output prediction intervals or not

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        predictions = []

        for t in range(0,h):
            data1 = self.data_original.iloc[0:-h+t,:]
            data2 = self.data_original.iloc[-h+t:,:]
            x = DynReg(formula=self.formula, data=data1)

            if fit_once is False:
                x.fit(printer=False, fit_method=fit_method)
            if t == 0:
                if fit_once is True:
                    x.fit(printer=False, fit_method=fit_method)
                    saved_lvs = x.latent_variables
                predictions = x.predict(1, oos_data=data2, intervals=intervals)
            else:
                if fit_once is True:
                    x.latent_variables = saved_lvs
                predictions = pd.concat([predictions,x.predict(h=1, oos_data=data2, intervals=intervals)])
        
        predictions.rename(columns={0:self.y_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions

    def plot_predict_is(self, h=5, fit_once=True, fit_method='MLE', **kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        fit_once : boolean
            Whether to fit the model once at the beginning (True), or fit every iteration (False)

        fit_method : string
            Which method to fit the model with

        Returns
        ----------
        - Plot of the forecast against data 
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        predictions = self.predict_is(h=h, fit_once=fit_once, fit_method=fit_method)
        data = self.data[-h:]
        plt.plot(predictions.index,data,label='Data')
        plt.plot(predictions.index,predictions,label='Predictions',c='black')
        plt.title(self.y_name)
        plt.legend(loc=2)   
        plt.show()          

    def simulation_smoother(self,beta):
        """ Koopman's simulation smoother - simulates from states given
        model parameters and observations

        Parameters
        ----------

        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        - A simulated state evolution
        """         

        T, Z, R, Q, H = self._ss_matrices(beta)

        # Generate e_t+ and n_t+
        rnd_h = np.random.normal(0,np.sqrt(H),self.data.shape[0]+1)
        q_dist = ss.multivariate_normal([0.0, 0.0], Q)
        rnd_q = q_dist.rvs(self.data.shape[0]+1)

        # Generate a_t+ and y_t+
        a_plus = np.zeros((T.shape[0],self.data.shape[0]+1)) 
        a_plus[0,0] = np.mean(self.data[0:5])
        y_plus = np.zeros(self.data.shape[0])

        for t in range(0,self.data.shape[0]+1):
            if t == 0:
                a_plus[:,t] = np.dot(T,a_plus[:,t]) + rnd_q[t,:]
                y_plus[t] = np.dot(Z,a_plus[:,t]) + rnd_h[t]
            else:
                if t != self.data.shape[0]:
                    a_plus[:,t] = np.dot(T,a_plus[:,t-1]) + rnd_q[t,:]
                    y_plus[t] = np.dot(Z,a_plus[:,t]) + rnd_h[t]

        alpha_hat, _ = self.smoothed_state(self.data,beta)
        alpha_hat_plus, _ = self.smoothed_state(y_plus,beta)
        alpha_tilde = alpha_hat - alpha_hat_plus + a_plus
    
        return alpha_tilde

    def smoothed_state(self,data,beta):
        """ Creates the negative log marginal likelihood of the model

        Parameters
        ----------

        data : np.array
            Data to be smoothed

        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        - Smoothed states
        """         

        T, Z, R, Q, H = self._ss_matrices(beta)
        alpha, V = dl_univariate_KFS(data,Z,H,T,Q,R,0.0)
        return alpha, V

    def sample(self, nsims=1000):
        """ Samples from the posterior predictive distribution

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws from the posterior predictive distribution

        Returns
        ----------
        - np.ndarray of draws from the data
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            lv_draws = self.draw_latent_variables(nsims=nsims)

            mus = []
            for i in range(nsims):
                mu, V = self.smoothed_state(self.data, lv_draws[:,i])
                _, Z, _, _, _ = self._ss_matrices(lv_draws[:,i])
                smoothed_series = np.zeros(self.y.shape[0])
                for t in range(0,self.y.shape[0]):
                    smoothed_series[t] = np.dot(Z[t],mu[:,t])
                mus.append(smoothed_series)

            data_draws = np.array([np.random.normal(mus[i], np.sqrt(self.latent_variables.z_list[0].prior.transform(lv_draws[0,i])), mus[i].shape[0]) for i in range(nsims)])
            return data_draws

    def plot_sample(self, nsims=10, plot_data=True, **kwargs):
        """
        Plots draws from the posterior predictive density against the data

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws from the posterior predictive distribution

        plot_data boolean
            Whether to plot the data or not
        """

        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            import matplotlib.pyplot as plt
            import seaborn as sns

            figsize = kwargs.get('figsize',(10,7))
            plt.figure(figsize=figsize)
            date_index = self.index
            draws = self.sample(nsims).T
            plt.plot(date_index, draws, label='Posterior Draws', alpha=1.0)
            if plot_data is True:
                plt.plot(date_index, self.data, label='Data', c='black', alpha=0.5, linestyle='', marker='s')
            plt.title(self.data_name)
            plt.show()    

    def ppc(self, nsims=1000, T=np.mean):
        """ Computes posterior predictive p-value

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws for the PPC

        T : function
            A discrepancy measure - e.g. np.mean, np.std, np.max

        Returns
        ----------
        - float (posterior predictive p-value)
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            lv_draws = self.draw_latent_variables(nsims=nsims)
            mus = []
            for i in range(nsims):
                mu, V = self.smoothed_state(self.data, lv_draws[:,i])
                _, Z, _, _, _ = self._ss_matrices(lv_draws[:,i])
                smoothed_series = np.zeros(self.y.shape[0])
                for t in range(0,self.y.shape[0]):
                    smoothed_series[t] = np.dot(Z[t],mu[:,t])
                mus.append(smoothed_series)
            data_draws = np.array([np.random.normal(mus[i], np.sqrt(self.latent_variables.z_list[0].prior.transform(lv_draws[0,i])), mus[i].shape[0]) for i in range(nsims)])
            T_sims = T(self.sample(nsims=nsims), axis=1)
            T_actual = T(self.data)
            return len(T_sims[T_sims>T_actual])/nsims

    def plot_ppc(self, nsims=1000, T=np.mean, **kwargs):
        """ Plots histogram of the discrepancy from draws of the posterior

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws for the PPC

        T : function
            A discrepancy measure - e.g. np.mean, np.std, np.max
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            import matplotlib.pyplot as plt
            import seaborn as sns

            figsize = kwargs.get('figsize',(10,7))

            lv_draws = self.draw_latent_variables(nsims=nsims)
            mus = []
            for i in range(nsims):
                mu, V = self.smoothed_state(self.data, lv_draws[:,i])
                _, Z, _, _, _ = self._ss_matrices(lv_draws[:,i])
                smoothed_series = np.zeros(self.y.shape[0])
                for t in range(0,self.y.shape[0]):
                    smoothed_series[t] = np.dot(Z[t],mu[:,t])
                mus.append(smoothed_series)
            data_draws = np.array([np.random.normal(mus[i], np.sqrt(self.latent_variables.z_list[0].prior.transform(lv_draws[0,i])), mus[i].shape[0]) for i in range(nsims)])
            T_sim = T(self.sample(nsims=nsims), axis=1)
            T_actual = T(self.data)

            if T == np.mean:
                description = " of the mean"
            elif T == np.max:
                description = " of the maximum"
            elif T == np.min:
                description = " of the minimum"
            elif T == np.median:
                description = " of the median"
            else:
                description = ""

            plt.figure(figsize=figsize)
            ax = plt.subplot()
            ax.axvline(T_actual)
            sns.distplot(T_sim, kde=False, ax=ax)
            ax.set(title='Posterior predictive' + description, xlabel='T(x)', ylabel='Frequency');
            plt.show()


import numpy as np
import scipy.stats as ss
import scipy.special as sp

from .family import Family
from .flat import Flat
from .normal import Normal

from .gas_recursions import gas_recursion_skewt_orderone, gas_recursion_skewt_ordertwo
from .gas_recursions import gasx_recursion_skewt_orderone, gasx_recursion_skewt_ordertwo
from .gas_recursions import gas_llev_recursion_skewt_orderone, gas_llev_recursion_skewt_ordertwo
from .gas_recursions import gas_llt_recursion_skewt_orderone, gas_llt_recursion_skewt_ordertwo
from .gas_recursions import gas_reg_recursion_skewt_orderone, gas_reg_recursion_skewt_ordertwo


class Skewt(Family):
    """ 
    Student Skew t Distribution
    ----
    This class contains methods relating to the Student Skew t distribution for time series.
    """

    def __init__(self, loc=0.0, scale=1.0, df=8.0, gamma=1.0, transform=None, **kwargs):
        """
        Parameters
        ----------
        loc : float
            Location parameter for the Skew t distribution

        scale : float
            Scale parameter for the Skew t distribution

        df : float
            Degrees of freedom parameter for the Skew t distribution

        gamma : float
            Skewness parameter (1.0 is skewed; under 1.0, -ve skewed; over 1.0, +ve skewed)

        transform : str
            Whether to apply a transformation to the location variable - e.g. 'exp' or 'logit'
        """
        super(Skewt, self).__init__(transform)
        self.loc0 = loc
        self.scale0 = scale
        self.df0 = df
        self.gamma0 = gamma
        self.covariance_prior = False

        self.gradient_only = kwargs.get('gradient_only', False) # used for GAS t models
        if self.gradient_only is True:
            self.score_function = self.first_order_score
        else:
            self.score_function = self.second_order_score

    def approximating_model(self, beta, T, Z, R, Q, h_approx, data):
        """ Creates approximating Gaussian state space model for Skewt measurement density
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm
        
        h_approx : float
            The variance of the measurement density
        
        data: np.array
            The univariate time series data

        Returns
        ----------
        H : np.array
            Approximating measurement variance matrix
        
        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(data.shape[0])*h_approx
        mu = np.zeros(data.shape[0])

        return H, mu

    def approximating_model_reg(self, beta, T, Z, R, Q, h_approx, data, X, state_no):
        """ Creates approximating Gaussian state space model for Skewt measurement density
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm
        
        h_approx : float
            The variance of the measurement density
        
        data: np.array
            The univariate time series data

        X: np.array
            The regressors

        state_no : int
            Number of states
    
        Returns
        ----------
        H : np.array
            Approximating measurement variance matrix
        
        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(data.shape[0])*h_approx
        mu = np.zeros(data.shape[0])

        return H, mu

    @staticmethod
    def build_latent_variables():
        """ Builds additional latent variables for this family

        Returns
        ----------
        - A list of lists (each sub-list contains latent variable information)
        """
        lvs_to_build = []
        lvs_to_build.append(['Skewness', Flat(transform='exp'), Normal(0, 3), 0.0])
        lvs_to_build.append(['Skewt Scale', Flat(transform='exp'), Normal(0, 3), 0.01])
        lvs_to_build.append(['v', Flat(transform='exp'), Normal(0, 3), 2.5])
        return lvs_to_build

    @staticmethod
    def draw_variable(loc, scale, shape, skewness, nsims):
        """ Draws random variables from Skew t distribution

        Parameters
        ----------
        loc : float
            location parameter for the distribution

        scale : float
            scale parameter for the distribution

        shape : float
            tail thickness parameter for the distribution

        skewness : float
            skewness parameter for the distribution

        nsims : int or list
            number of draws to take from the distribution

        Returns
        ----------
        - Random draws from the distribution
        """
        return loc + scale*Skewt.rvs(shape, skewness, nsims)

    @staticmethod
    def first_order_score(y, mean, scale, shape, skewness):
        """ GAS Skew t Update term using gradient only - native Python function

        Parameters
        ----------
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Skew t distribution

        scale : float
            scale parameter for the Skew t distribution

        shape : float
            tail thickness parameter for the Skew t distribution

        skewness : float
            skewness parameter for the Skew t distribution

        Returns
        ----------
        - Score of the Skew t family
        """
        m1 = (np.sqrt(shape)*sp.gamma((shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(shape/2.0))
        mean = mean + (skewness - (1.0/skewness))*scale*m1
        if (y-mean)>=0:
            return ((shape+1)/shape)*(y-mean)/(np.power(skewness*scale,2) + (np.power(y-mean,2)/shape))
        else:
            return ((shape+1)/shape)*(y-mean)/(np.power(scale,2) + (np.power(skewness*(y-mean),2)/shape))

    @staticmethod
    def rvs(df, gamma, n):
        """ Generates random variables from a Skew t distribution

        Parameters
        ----------  
        df : float
            degrees of freedom parameter

        gamma : float
            skewness parameter

        n : int or list
            Number of simulations to perform; if list input, produces array

        """

        if type(n) == list:
            u = np.random.uniform(size=n[0]*n[1])
            result = Skewt.ppf(q=u, df=df, gamma=gamma)
            result = np.split(result,n[0])
            return np.array(result)
        else:
            u = np.random.uniform(size=n)
            if isinstance(df, np.ndarray) or isinstance(gamma, np.ndarray):
                return np.array([Skewt.ppf(q=np.array([u[i]]), df=df[i], gamma=gamma[i])[0] for i in range(n)])
            else:
                return Skewt.ppf(q=u, df=df, gamma=gamma)

    @staticmethod
    def logpdf_internal(x, df, loc=0.0, scale=1.0, gamma = 1.0):
        result = np.zeros(x.shape[0])
        result[x-loc<0] = np.log(2.0) - np.log(gamma + 1.0/gamma) + ss.t.logpdf(x=gamma*x[(x-loc) < 0], loc=loc[(x-loc) < 0]*gamma,df=df, scale=scale)
        result[x-loc>=0] = np.log(2.0) - np.log(gamma + 1.0/gamma) + ss.t.logpdf(x=x[(x-loc) >= 0]/gamma, loc=loc[(x-loc) >= 0]/gamma,df=df, scale=scale)
        return result

    @staticmethod
    def logpdf_internal_prior(x, df, loc=0.0, scale=1.0, gamma = 1.0):
        if x-loc < 0.0:
            return np.log(2.0) - np.log(gamma + 1.0/gamma) + ss.t.logpdf(x=gamma*x, loc=loc*gamma,df=df, scale=scale)
        else:
            return np.log(2.0) - np.log(gamma + 1.0/gamma) + ss.t.logpdf(x=x/gamma, loc=loc/gamma,df=df, scale=scale)

    def logpdf(self, mu):
        """
        Log PDF for Skew t prior

        Parameters
        ----------
        mu : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - log(p(mu))
        """
        if self.transform is not None:
            mu = self.transform(mu)    
        return self.logpdf_internal_prior(mu, df=self.df0, loc=self.loc0, scale=self.scale0, gamma=self.gamma0) 

    @staticmethod
    def markov_blanket(y, mean, scale, shape, skewness):
        """ Markov blanket for each likelihood term

        Parameters
        ----------
        y : np.ndarray
            univariate time series

        mean : np.ndarray
            array of location parameters for the Skew t distribution

        scale : float
            scale parameter for the Skew t distribution

        shape : float
            tail thickness parameter for the Skew t distribution

        skewness : float
            skewness parameter for the Skew t distribution

        Returns
        ----------
        - Markov blanket of the Skew t family
        """
        m1 = (np.sqrt(shape)*sp.gamma((shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(shape/2.0))
        mean = mean + (skewness - (1.0/skewness))*scale*m1
        return Skewt.logpdf_internal(x=y, df=shape, loc=mean, gamma=skewness, scale=scale)

    @staticmethod
    def setup():
        """ Returns the attributes of this family

        Notes
        ----------
        - scale notes whether family has a variance parameter (sigma)
        - shape notes whether family has a tail thickness parameter (nu)
        - skewness notes whether family has a skewness parameter (gamma)
        - mean_transform is a function which transforms the location parameter
        - cythonized notes whether the family has cythonized routines
        
        Returns
        ----------
        - model name, link function, scale, shape, skewness, mean_transform, cythonized
        """
        name = "Skewt"
        link = np.array
        scale = True
        shape = True
        skewness = True
        mean_transform = np.array
        cythonized = True
        return name, link, scale, shape, skewness, mean_transform, cythonized

    @staticmethod
    def neg_loglikelihood(y, mean, scale, shape, skewness):
        """ Negative loglikelihood function

        Parameters
        ----------
        y : np.ndarray
            univariate time series

        mean : np.ndarray
            array of location parameters for the Skew t distribution

        scale : float
            scale parameter for the Skew t distribution

        shape : float
            tail thickness parameter for the Skew t distribution

        skewness : float
            skewness parameter for the Skew t distribution

        Returns
        ----------
        - Negative loglikelihood of the Skew t family
        """
        m1 = (np.sqrt(shape)*sp.gamma((shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(shape/2.0))
        mean = mean + (skewness - (1.0/skewness))*scale*m1
        return -np.sum(Skewt.logpdf_internal(x=y, df=shape, loc=mean, gamma=skewness, scale=scale))

    @staticmethod
    def pdf_internal(x, df, loc=0.0, scale=1.0, gamma = 1.0):
        """
        Raw PDF function for the Skew t distribution
        """
        result = np.zeros(x.shape[0])
        result[x<0] = 2.0/(gamma + 1.0/gamma)*stats.t.pdf(x=gamma*x[(x-loc) < 0], loc=loc[(x-loc) < 0]*gamma,df=df, scale=scale)
        result[x>=0] = 2.0/(gamma + 1.0/gamma)*stats.t.pdf(x=x[(x-loc) >= 0]/gamma, loc=loc[(x-loc) >= 0]/gamma,df=df, scale=scale)
        return result

    def pdf(self, mu):
        """
        PDF for Skew t prior

        Parameters
        ----------
        mu : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - p(mu)
        """
        if self.transform is not None:
            mu = self.transform(mu)    
        return self.pdf_internal(mu, df=self.df0, loc=self.loc0, scale=self.scale0, gamma=self.gamma0) 

    @staticmethod
    def reg_score_function(X, y, mean, scale, shape, skewness):
        """ GAS Skew t Regression Update term using gradient only - native Python function

        Parameters
        ----------
        X : float
            datapoint for the right hand side variable
    
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Skew t distribution

        scale : float
            scale parameter for the Skew t distribution

        shape : float
            tail thickness parameter for the Skew t distribution

        skewness : float
            skewness parameter for the Skew t distribution

        Returns
        ----------
        - Score of the Skew t family
        """
        m1 = (np.sqrt(shape)*sp.gamma((shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(shape/2.0))
        mean = mean + (skewness - (1.0/skewness))*scale*m1
        if (y-mean)>=0:
            return ((shape+1)/shape)*((y-mean)*X)/(np.power(skewness*scale,2) + (np.power(y-mean,2)/shape))
        else:
            return ((shape+1)/shape)*((y-mean)*X)/(np.power(scale,2) + (np.power(skewness*(y-mean),2)/shape))

    @staticmethod
    def second_order_score(y, mean, scale, shape, skewness):
        """ GAS Skew t Update term potentially using second-order information - native Python function

        Parameters
        ----------
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Skew t distribution

        scale : float
            scale parameter for the Skew t distribution

        shape : float
            tail thickness parameter for the Skew t distribution

        skewness : float
            skewness parameter for the Skew t distribution

        Returns
        ----------
        - Adjusted score of the Skew t family
        """
        m1 = (np.sqrt(shape)*sp.gamma((shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(shape/2.0))
        mean = mean + (skewness - (1.0/skewness))*scale*m1
        if (y-mean)>=0:
            return ((shape+1)/shape)*(y-mean)/(np.power(skewness*scale,2) + (np.power(y-mean,2)/shape))
        else:
            return ((shape+1)/shape)*(y-mean)/(np.power(scale,2) + (np.power(skewness*(y-mean),2)/shape))

    @staticmethod
    def cdf(x, df, loc=0.0, scale=1.0, gamma = 1.0):
        """
        CDF function for Skew t distribution
        """
        result = np.zeros(x.shape[0])
        result[x<0] = 2.0/(np.power(gamma,2) + 1.0)*ss.t.cdf(gamma*(x[x-loc < 0]-loc[x-loc < 0])/scale, df=df)
        result[x>=0] = 1.0/(np.power(gamma,2) + 1.0) + 2.0/((1.0/np.power(gamma,2)) + 1.0)*(ss.t.cdf((x[x-loc >= 0]-loc[x-loc >= 0])/(gamma*scale), df=df)-0.5)
        return result

    @staticmethod
    def ppf(q, df, loc=0.0, scale=1.0, gamma = 1.0):
        """
        PPF function for Skew t distribution
        """
        result = np.zeros(q.shape[0])
        probzero = Skewt.cdf(x=np.zeros(1),loc=np.zeros(1),df=df,gamma=gamma)
        result[q<probzero] = 1.0/gamma*ss.t.ppf(((np.power(gamma,2) + 1.0) * q[q<probzero])/2.0,df)
        result[q>=probzero] = gamma*ss.t.ppf((1.0 + 1.0/np.power(gamma,2))/2.0*(q[q >= probzero] - probzero) + 0.5, df)
        return result

    # Optional Cythonized recursions below for GAS Skew t models

    @staticmethod
    def gradient_recursion():
        """ GAS Skew t Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Skew t model - gradient only
        """
        return gas_recursion_skewt_orderone

    @staticmethod
    def newton_recursion():
        """ GAS Skew t Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Skew t model - adjusted score
        """
        return gas_recursion_skewt_ordertwo

    @staticmethod
    def gradientx_recursion():
        """ GASX Skew t Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GASX Skew t model - gradient only
        """
        return gasx_recursion_skewt_orderone

    @staticmethod
    def newtonx_recursion():
        """ GASX Skew t Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GASX Skew t model - adjusted score
        """
        return gasx_recursion_skewt_ordertwo

    @staticmethod
    def gradientllev_recursion():
        """ GAS Local Level Skew t Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Local Level Skew t model - gradient only
        """
        return gas_llev_recursion_skewt_orderone

    @staticmethod
    def newtonllev_recursion():
        """ GAS Local Level Skew t Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Local Level Skew t model - adjusted score
        """
        return gas_llev_recursion_skewt_ordertwo

    @staticmethod
    def gradientllt_recursion():
        """ GAS Local Linear Trend Skew t Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Local Linear Trend Skew t model - gradient only
        """
        return gas_llt_recursion_skewt_orderone

    @staticmethod
    def newtonllt_recursion():
        """ GAS Local Linear Trend Skew t Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Local Linear Trend Skew t model - adjusted score
        """
        return gas_llt_recursion_skewt_ordertwo

    @staticmethod
    def gradientreg_recursion():
        """ GAS Dynamic Regression Skew t Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Dynamic Regression Skew t model - gradient only
        """
        return gas_reg_recursion_skewt_orderone

    @staticmethod
    def newtonreg_recursion():
        """ GAS Dynamic Regression Skew t Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Dynamic Regression Skew t model - adjusted score
        """
        return gas_reg_recursion_skewt_ordertwo



import sys
import numpy as np

from .. import families as fam

from .kernel_routines import SE_K_matrix, SE_Kstar_matrix, SE_Kstarstar_matrix, SE_K_arbitrary_X_matrix
from .kernel_routines import OU_K_matrix, OU_Kstar_matrix, OU_Kstarstar_matrix, OU_K_arbitrary_X_matrix
from .kernel_routines import RQ_K_matrix, RQ_Kstar_matrix, RQ_Kstarstar_matrix, RQ_K_arbitrary_X_matrix
from .kernel_routines import ARD_K_matrix, ARD_Kstar_matrix, ARD_Kstarstar_matrix, ARD_K_arbitrary_X_matrix
from .kernel_routines import Periodic_K_matrix, Periodic_Kstar_matrix, Periodic_Kstarstar_matrix, Periodic_K_arbitrary_X_matrix


class SquaredExponential(object):
    """ Squared Exponential Kernel

    Parameters
    ----------
    X : np.ndarray
        The RHS factors for the GP regression
    """

    def __init__(self, X=np.array([1])):
        self.X = X.transpose()

    @staticmethod
    def build_latent_variables():
        """ Builds latent variables for this kernel

        Returns
        ----------
        - A list of lists (each sub-list contains latent variable information)
        """
        lvs_to_build = []
        lvs_to_build.append(['Noise Sigma^2', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        lvs_to_build.append(['l', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        lvs_to_build.append(['tau', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        return lvs_to_build

    def K(self, parm):
        return SE_K_matrix(self.X, parm) + np.identity(self.X.shape[0])*(10**-10)

    def K_arbitrary_X(self, parm, Xstar1, Xstar2):
        return SE_K_arbitrary_X_matrix(Xstar1, Xstar2, parm)

    def Kstar(self, parm, Xstar):
        return SE_Kstar_matrix(self.X, Xstar, parm)

    def Kstarstar(self, parm, Xstar):
        return SE_Kstarstar_matrix(Xstar, parm)


class OrnsteinUhlenbeck(object):
    """ Ornstein Uhlenbeck Kernel

    Parameters
    ----------
    X : np.ndarray
        The RHS factors for the GP regression
    """

    def __init__(self, X=np.array([1])):

        self.X = X.transpose()

    @staticmethod
    def build_latent_variables():
        """ Builds latent variables for this kernel

        Returns
        ----------
        - A list of lists (each sub-list contains latent variable information)
        """
        lvs_to_build = []
        lvs_to_build.append(['Noise Sigma^2', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        lvs_to_build.append(['l', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        lvs_to_build.append(['tau', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        return lvs_to_build

    def K(self, parm):
        """ Returns the Gram Matrix

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the Gram Matrix

        Returns
        ----------
        - Gram Matrix (np.ndarray)
        """
        return OU_K_matrix(self.X, parm) + np.identity(self.X.shape[0])*(10**-10)

    def K_arbitrary_X(self, parm, Xstar1, Xstar2):
        """ Returns K(x1,x2)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x1, x2)

        Xstar1 : np.ndarray
            First data subset

        Xstar2 : np.ndarray
            Second data subset

        Returns
        ----------
        - K(x1, x2)
        """
        return OU_K_arbitrary_X_matrix(Xstar1, Xstar2, parm)

    def Kstar(self, parm, Xstar):
        """ Returns K(x, x*)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x, x*)

        Xstar : np.ndarray
            Data for prediction

        Returns
        ----------
        - K(x, x*)
        """
        return OU_Kstar_matrix(self.X, Xstar, parm)

    def Kstarstar(self, parm, Xstar):
        """ Returns K(x*, x*)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x*, x*)

        Xstar : np.ndarray
            Data for prediction

        Returns
        ----------
        - K(x*, x*)
        """
        return OU_Kstarstar_matrix(Xstar, parm)


class ARD(object):
    """ ARD Kernel

    Parameters
    ----------
    X : np.ndarray
        The RHS factors for the GP regression
    """

    def __init__(self, X=np.array([1])):
        self.X = X.transpose()

    def build_latent_variables(self):
        """ Builds latent variables for this kernel

        Returns
        ----------
        - A list of lists (each sub-list contains latent variable information)
        """
        lvs_to_build = []
        lvs_to_build.append(['Noise Sigma^2', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        for lag in range(self.X.shape[1]):
            lvs_to_build.append(['l lag' + str(lag+1), fam.FLat(transform='exp'), fam.Normal(0,3), -1.0])
        lvs_to_build.append(['tau', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        return lvs_to_build

    def K(self, parm):
        """ Returns the Gram Matrix

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the Gram Matrix

        Returns
        ----------
        - Gram Matrix (np.ndarray)
        """
        return ARD_K_matrix(self.X, parm) + np.identity(self.X.shape[0])*(10**-10)

    def K_arbitrary_X(self, parm, Xstar1, Xstar2):
        """ Returns K(x1,x2)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x1, x2)

        Xstar1 : np.ndarray
            First data subset

        Xstar2 : np.ndarray
            Second data subset

        Returns
        ----------
        - K(x1, x2)
        """
        return ARD_K_arbitrary_X_matrix(Xstar1, Xstar2, parm)

    def Kstar(self, parm, Xstar):
        """ Returns K(x, x*)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x, x*)

        Xstar : np.ndarray
            Data for prediction

        Returns
        ----------
        - K(x, x*)
        """
        return ARD_Kstar_matrix(self.X, Xstar, parm)

    def Kstarstar(self, parm, Xstar):
        """ Returns K(x*, x*)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x*, x*)

        Xstar : np.ndarray
            Data for prediction

        Returns
        ----------
        - K(x*, x*)
        """
        return ARD_Kstarstar_matrix(Xstar, parm)


class RationalQuadratic(object):
    """ Rational Quadratic Kernel

    Parameters
    ----------
    X : np.ndarray
        The RHS factors for the GP regression
    """

    def __init__(self, X=np.array([1])):

        self.X = X.transpose()

    def build_latent_variables(self):
        lvs_to_build = []
        lvs_to_build.append(['Noise Sigma^2', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        lvs_to_build.append(['a', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        lvs_to_build.append(['l', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        lvs_to_build.append(['tau', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        return lvs_to_build

    def K(self, parm):
        """ Returns the Gram Matrix

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the Gram Matrix

        Returns
        ----------
        - Gram Matrix (np.ndarray)
        """
        return RQ_K_matrix(self.X, parm) + np.identity(self.X.shape[0])*(10**-10)

    def K_arbitrary_X(self, parm, Xstar1, Xstar2):
        """ Returns K(x1,x2)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x1, x2)

        Xstar1 : np.ndarray
            First data subset

        Xstar2 : np.ndarray
            Second data subset

        Returns
        ----------
        - K(x1, x2)
        """
        return RQ_K_arbitrary_X_matrix(Xstar1, Xstar2, parm)

    def Kstar(self, parm, Xstar):
        """ Returns K(x, x*)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x, x*)

        Xstar : np.ndarray
            Data for prediction

        Returns
        ----------
        - K(x, x*)
        """
        return RQ_Kstar_matrix(self.X, Xstar, parm)

    def Kstarstar(self, parm, Xstar):
        """ Returns K(x*, x*)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x*, x*)

        Xstar : np.ndarray
            Data for prediction
        
        Returns
        ----------
        - K(x*, x*)
        """
        return RQ_Kstarstar_matrix(Xstar, parm)


class Periodic(object):
    """ Periodic Kernel

    Parameters
    ----------
    X : np.ndarray
        The RHS factors for the GP regression
    """

    def __init__(self, X=np.array([1])):

        self.X = X.transpose()

    def build_latent_variables(self):
        """ Builds latent variables for this kernel

        Returns
        ----------
        - A list of lists (each sub-list contains latent variable information)
        """
        lvs_to_build = []
        lvs_to_build.append(['Noise Sigma^2', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        lvs_to_build.append(['l', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        lvs_to_build.append(['tau', fam.Flat(transform='exp'), fam.Normal(0,3), -1.0])
        return lvs_to_build

    def K(self, parm):
        """ Returns the Gram Matrix

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the Gram Matrix

        Returns
        ----------
        - Gram Matrix (np.ndarray)
        """
        return Periodic_K_matrix(self.X, parm) + np.identity(self.X.shape[0])*(10**-10)

    def K_arbitrary_X(self, parm, Xstar1, Xstar2):
        """ Returns K(x1,x2)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x1, x2)

        Xstar1 : np.ndarray
            First data subset

        Xstar2 : np.ndarray
            Second data subset

        Returns
        ----------
        - K(x1, x2)
        """
        return Periodic_K_arbitrary_X_matrix(Xstar1, Xstar2, parm)

    def Kstar(self, parm, Xstar):
        """ Returns K(x, x*)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x, x*)

        Xstar : np.ndarray
            Data for prediction

        Returns
        ----------
        - K(x, x*)
        """
        return Periodic_Kstar_matrix(self.X, Xstar, parm)

    def Kstarstar(self, parm, Xstar):
        """ Returns K(x*, x*)

        Parameters
        ----------
        parm : np.ndarray
            Parameters for the K(x*, x*)

        Xstar : np.ndarray
            Data for prediction

        Returns
        ----------
        - K(x*, x*)
        """
        return Periodic_Kstarstar_matrix(Xstar, parm)


import numpy as np
import scipy.stats as ss
import scipy.special as sp

from .family import Family
from .flat import Flat
from .normal import Normal

from .gas_recursions import gas_recursion_exponential_orderone, gas_recursion_exponential_ordertwo
from .gas_recursions import gasx_recursion_exponential_orderone, gasx_recursion_exponential_ordertwo
from .gas_recursions import gas_llev_recursion_exponential_orderone, gas_llev_recursion_exponential_ordertwo
from .gas_recursions import gas_llt_recursion_exponential_orderone, gas_llt_recursion_exponential_ordertwo
from .gas_recursions import gas_reg_recursion_exponential_orderone, gas_reg_recursion_exponential_ordertwo


class Exponential(Family):
    """ 
    Exponential Distribution
    ----
    This class contains methods relating to the Exponential distribution for time series.
    """

    def __init__(self, lmd=1.0, transform=None, **kwargs):
        """
        Parameters
        ----------
        lambda : float
            Rate parameter for the Exponential distribution

        transform : str
            Whether to apply a transformation to the location variable - e.g. 'exp' or 'logit'
        """
        super(Exponential, self).__init__(transform)
        self.lmd0 = lmd
        self.covariance_prior = False

        self.gradient_only = kwargs.get('gradient_only', False) # used for GAS Exponential models
        if self.gradient_only is True:
            self.score_function = self.first_order_score
        else:
            self.score_function = self.second_order_score

    def approximating_model(self, beta, T, Z, R, Q, h_approx, data):
        """ Creates approximating Gaussian state space model for Exponential measurement density
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm
        
        h_approx : float
            The variance of the measurement density

        data: np.array
            The univariate time series data
        
        Returns
        ----------
        H : np.array
            Approximating measurement variance matrix
        
        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(data.shape[0])*h_approx
        mu = np.zeros(data.shape[0])

        return H, mu


    def approximating_model_reg(self, beta, T, Z, R, Q, h_approx, data, X, state_no):
        """ Creates approximating Gaussian state space model for Exponential measurement density
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm
        
        h_approx : float
            The variance of the measurement density

        data: np.array
            The univariate time series data
        
        X: np.array
            The regressors

        state_no : int
            Number of states

        Returns
        ----------
        H : np.array
            Approximating measurement variance matrix
        
        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(data.shape[0])*h_approx
        mu = np.zeros(data.shape[0])

        return H, mu

    @staticmethod
    def build_latent_variables():
        """ Builds additional latent variables for this family

        Returns
        ----------
        - A list of lists (each sub-list contains latent variable information)
        """
        lvs_to_build = []
        return lvs_to_build

    @staticmethod
    def draw_variable(loc, scale, shape, skewness, nsims):
        """ Draws random variables from Exponential distribution        

        Parameters
        ----------
        loc : float
            location parameter for the distribution

        scale : float
            scale parameter for the distribution

        shape : float
            tail thickness parameter for the distribution

        skewness : float
            skewness parameter for the distribution

        nsims : int or list
            number of draws to take from the distribution

        Returns
        ----------
        - Random draws from the distribution
        """
        return np.random.exponential(1.0/loc, nsims)

    @staticmethod
    def first_order_score(y, mean, scale, shape, skewness):
        """ GAS Exponential Update term using gradient only - native Python function

        Parameters
        ----------
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Exponential distribution

        scale : float
            scale parameter for the Exponential distribution

        shape : float
            tail thickness parameter for the Exponential distribution

        skewness : float
            skewness parameter for the Exponential distribution

        Returns
        ----------
        - Score of the Exponential family
        """
        return 1 - (mean*y)

    def logpdf(self, mu):
        """
        Log PDF for Exponential prior

        Parameters
        ----------
        mu : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - log(p(mu))
        """
        if self.transform is not None:
            mu = self.transform(mu)    
        return ss.expon.logpdf(mu, self.lmd0) 

    @staticmethod
    def markov_blanket(y, mean, scale, shape, skewness):
        """ Markov blanket for the Exponential distribution

        Parameters
        ----------
        y : np.ndarray
            univariate time series

        mean : np.ndarray
            array of location parameters for the Exponential distribution

        scale : float
            scale parameter for the Exponential distribution

        shape : float
            tail thickness parameter for the Exponential distribution

        skewness : float
            skewness parameter for the Exponential distribution

        Returns
        ----------
        - Markov blanket of the Exponential family
        """
        return ss.expon.logpdf(x=y, scale=1/mean)

    @staticmethod
    def exponential_link(x):
        return 1.0/np.exp(x)

    @staticmethod
    def setup():
        """ Returns the attributes of this family

        Notes
        ----------
        - scale notes whether family has a variance parameter (sigma)
        - shape notes whether family has a tail thickness parameter (nu)
        - skewness notes whether family has a skewness parameter (gamma)
        - mean_transform is a function which transforms the location parameter
        - cythonized notes whether the family has cythonized routines
        
        Returns
        ----------
        - model name, link function, scale, shape, skewness, mean_transform, cythonized
        """
        name = "Exponential GAS"
        link = Exponential.exponential_link
        scale = False
        shape = False
        skewness = False
        mean_transform = np.log
        cythonized = True
        return name, link, scale, shape, skewness, mean_transform, cythonized

    @staticmethod
    def neg_loglikelihood(y, mean, scale, shape, skewness):
        """ Negative loglikelihood function

        Parameters
        ----------
        y : np.ndarray
            univariate time series

        mean : np.ndarray
            array of location parameters for the Exponential distribution

        scale : float
            scale parameter for the Exponential distribution

        shape : float
            tail thickness parameter for the Exponential distribution

        skewness : float
            skewness parameter for the Exponential distribution

        Returns
        ----------
        - Negative loglikelihood of the Exponential family
        """
        return -np.sum(ss.expon.logpdf(x=y, scale=1/mean))

    def pdf(self, mu):
        """
        PDF for Exponential prior

        Parameters
        ----------
        mu : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - p(mu)
        """
        if self.transform is not None:
            mu = self.transform(mu)                
        return ss.expon.pdf(mu, self.lmd0) 

    @staticmethod
    def reg_score_function(X, y, mean, scale, shape, skewness):
        """ GAS Exponential Regression Update term using gradient only - native Python function

        Parameters
        ----------
        X : float
            datapoint for the right hand side variable
    
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Exponential distribution

        scale : float
            scale parameter for the Exponential distribution

        shape : float
            tail thickness parameter for the Exponential distribution

        skewness : float
            skewness parameter for the Exponential distribution

        Returns
        ----------
        - Score of the Exponential family
        """
        return X*(1.0 - mean*y)

    @staticmethod
    def second_order_score(y, mean, scale, shape, skewness):
        """ GAS Exponential Update term potentially using second-order information - native Python function

        Parameters
        ----------
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Exponential distribution

        scale : float
            scale parameter for the Exponential distribution

        shape : float
            tail thickness parameter for the Exponential distribution

        skewness : float
            skewness parameter for the Exponential distribution

        Returns
        ----------
        - Adjusted score of the Exponential family
        """
        return 1 - (mean*y)

    # Optional Cythonized recursions below for GAS Exponential models

    @staticmethod
    def gradient_recursion():
        """ GAS Exponential Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Exponential model - gradient only
        """
        return gas_recursion_exponential_orderone

    @staticmethod
    def newton_recursion():
        """ GAS Exponential Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Exponential model - adjusted score
        """
        return gas_recursion_exponential_ordertwo

    @staticmethod
    def gradientx_recursion():
        """ GASX Exponential Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GASX Exponential model - gradient only
        """
        return gasx_recursion_exponential_orderone

    @staticmethod
    def newtonx_recursion():
        """ GASX Exponential Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GASX Exponential model - adjusted score
        """
        return gasx_recursion_exponential_ordertwo

    @staticmethod
    def gradientllev_recursion():
        """ GAS Local Level Exponential Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Local Level Exponential model - gradient only
        """
        return gas_llev_recursion_exponential_orderone

    @staticmethod
    def newtonllev_recursion():
        """ GAS Local Level Exponential Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Local Level Exponential model - adjusted score
        """
        return gas_llev_recursion_exponential_ordertwo

    @staticmethod
    def gradientllt_recursion():
        """ GAS Local Linear Trend Exponential Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Local Linear Trend Exponential model - gradient only
        """
        return gas_llt_recursion_exponential_orderone

    @staticmethod
    def newtonllt_recursion():
        """ GAS Local Linear Trend Exponential Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Local Linear Trend Exponential model - adjusted score
        """
        return gas_llt_recursion_exponential_ordertwo

    @staticmethod
    def gradientreg_recursion():
        """ GAS Dynamic Regression Exponential Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Dynamic Regression Exponential model - gradient only
        """
        return gas_reg_recursion_exponential_orderone

    @staticmethod
    def newtonreg_recursion():
        """ GAS Dynamic Regression Exponential Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Dynamic Regression Exponential model - adjusted score
        """
        return gas_reg_recursion_exponential_ordertwo


import numpy as np

class Family(object):

    def __init__(self, transform=None, **kwargs):
        """
        Parameters
        ----------
        transform : str
            Whether to apply a transformation - e.g. 'exp' or 'logit'
        """
        self.transform_name = transform     
        self.transform = self.transform_define(transform)
        self.itransform = self.itransform_define(transform)
        self.itransform_name = self.itransform_name_define(transform)

    @staticmethod
    def ilogit(x):
        return 1.0/(1.0+np.exp(-x))

    @staticmethod
    def logit(x):
        return np.log(x) - np.log(1.0 - x)

    @staticmethod
    def transform_define(transform):
        """
        This function links the user's choice of transformation with the associated numpy function
        """
        if transform == 'tanh':
            return np.tanh
        elif transform == 'exp':
            return np.exp
        elif transform == 'logit':
            return Family.ilogit
        elif transform is None:
            return np.array
        else:
            return None

    @staticmethod
    def itransform_define(transform):
        """
        This function links the user's choice of transformation with its inverse
        """
        if transform == 'tanh':
            return np.arctanh
        elif transform == 'exp':
            return np.log
        elif transform == 'logit':
            return Family.logit
        elif transform is None:
            return np.array
        else:
            return None

    @staticmethod
    def itransform_name_define(transform):
        """
        This function is used for model results table, displaying any transformations performed
        """
        if transform == 'tanh':
            return 'arctanh'
        elif transform == 'exp':
            return 'log'
        elif transform == 'logit':
            return 'ilogit'
        elif transform is None:
            return ''
        else:
            return None
